#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""
Created on Thu Nov  8 17:49:54 2018

@author:
Maximilian N. Günther
MIT Kavli Institute for Astrophysics and Space Research, 
Massachusetts Institute of Technology,
77 Massachusetts Avenue,
Cambridge, MA 02109, 
USA
Email: maxgue@mit.edu
Web: www.mnguenther.com
"""

from __future__ import print_function, division, absolute_import

#::: plotting settings
import seaborn as sns
sns.set(context='paper', style='ticks', palette='deep', font='sans-serif', font_scale=1.5, color_codes=True)
sns.set_style({"xtick.direction": "in","ytick.direction": "in"})
sns.set_context(rc={'lines.markeredgewidth': 1})

#::: modules
import numpy as np
import matplotlib.pyplot as plt




def get_normalized_flux_from_normalized_mag( normalized_mag, normalized_mag_err=None ):
    '''
    Inputs:
    -------
    
    normalized_mag : float or array of float
        the normalized magnitude (i.e. centered around 0)
        
    normalized_mag_err : float or array of float (optional; default is None)
        the error on the normalized magnitude
        if not given, only the normalized_flux is returned
        if given, both the normalized_flux and the normalized_flux_err are returned
        
    
    Returns:
    --------
    
    normalized_flux : float or array of float
        the normalized_flux
        
    normalized_flux_err: float or array of float
        the error on the normalized_flux
    '''
    if normalized_mag_err is None:
        normalized_flux = 10.**(- normalized_mag/2.5 )
        return normalized_flux
    
    else:
        normalized_flux = 10.**(- normalized_mag/2.5 )
        conv = 2.5/np.log(10)
        normalized_flux_err = (normalized_mag_err/conv)*normalized_flux
        return normalized_flux, normalized_flux_err



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed May  8 18:41:19 2019

@author:
Maximilian N. Günther
MIT Kavli Institute for Astrophysics and Space Research, 
Massachusetts Institute of Technology,
77 Massachusetts Avenue,
Cambridge, MA 02109, 
USA
Email: maxgue@mit.edu
Web: www.mnguenther.com
"""

from __future__ import print_function, division, absolute_import

#::: modules
import numpy as np
import matplotlib.pyplot as plt
import os
from tqdm import tqdm
import itertools
import wotan
from transitleastsquares import period_grid as tls_period_grid
# from transitleastsquares import catalog_info
from astropy.constants import G
from astropy import units as u
import ellc
import time as timer
import multiprocessing
multiprocessing.set_start_method('forkserver', force=True)
from multiprocessing import cpu_count
from pathos.multiprocessing import ProcessingPool as Pool
from contextlib import closing

#::: my modules
from ..lightcurves.lightcurves import index_eclipses_smart
from .transit_search import get_tls_kwargs_by_tic, tls_search
from .injection_recovery_output import irplot
try:
    from exoworlds.tess import tessio
except:
    pass
from ..v2.classes import allesclass2
from ..v2.translator import translate

#::: plotting settings
import seaborn as sns
sns.set(context='paper', style='ticks', palette='deep', font='sans-serif', font_scale=1.5, color_codes=True)
sns.set_style({"xtick.direction": "in","ytick.direction": "in"})
sns.set_context(rc={'lines.markeredgewidth': 1})

#::: settings and constants
np.random.seed(42)
eps = 1e-12




###############################################################################
#::: helper functiomns
###############################################################################

#::: set up the logfile
def setup_logfile(options):
    '''
    Set up a logfile for the injection-recovery tests;
    starts user dialog if a logfile already exists
    
    Inputs:
    -------
    logfname : str
        file path and name for the logfile
    
    Returns:
    -------
    ex : np struct array / None
        existing logfile
    '''
    f = os.path.join(options['outdir'],options['logfname'])
    
    if os.path.exists(f):
        if options['user_response'] is None:
            response = input('Log file already exists. Do you want to (1) overwrite it, (2) append missing rows, or (3) abort?\n')
        else:
            response = options['user_response']
        if response == '1':   
            print('Overwriting log file...')
            with open(f,'w') as f: 
                f.write('inj_period,inj_rplanet,inj_depth,inj_epoch,tls_period,tls_depth,tls_duration,tls_epoch,tls_SNR,tls_SDE,tls_FAP\n')
            return None
        elif response == '2': 
            print('Appending missing rows to log file...')
            ex = np.genfromtxt(f, names=True, dtype=None, delimiter=',')
            return ex
        else:
            raise ValueError('User aborted. Response was: "'+str(response)+'"')
    else:
        with open(f,'w') as f: 
            f.write('inj_period,inj_rplanet,inj_depth,inj_epoch,tls_period,tls_depth,tls_duration,tls_epoch,tls_SNR,tls_SDE,tls_FAP\n')
        return None
                


#round and str to avoid float issues
def mystr(x):
    return '{:.3f}'.format(x)


#::: check if a certain injection has already been logged, or if it needs to be done
def to_do_or_not_to_do_that_is_the_question(ex, period, rplanet):
    #if no previous logfile exists, then do it
    if ex is None:
        return True #do it
    #if a previous logfile exists and it includes those period and rplanet values (to 1e-6 precision), then skip this pair
    else:
        N = len(np.atleast_1d(ex['inj_period']))
        for i in range(N):
            if (np.abs(period - np.atleast_1d(ex['inj_period'])[i]) < 1e-6) and (np.abs(rplanet - np.atleast_1d(ex['inj_rplanet'])[i]) < 1e-6):
                return False #skip it
    #if the pair was not skipped before, then go ahead and do it
    return True #do it
    


# def inject(time, flux, flux_err, epoch, period, r_companion_earth, r_host=1, m_host=1, ldc=[0.5,0.5], window=None):
#     dic = translate(period=period,
#                     r_companion_earth=r_companion_earth,
#                     m_companion_earth=0,
#                     r_host=r_host,
#                     m_host=m_host, 
#                     quiet=True)
#     alles = allesclass2()
#     alles.settings = {'companions_phot':['b'], 'inst_phot':['telescope'], 'host_ld_law_telescope':'quad'}
#     alles.params = {'b_rr':dic['rr'], 'b_rsuma':dic['rsuma'], 'b_epoch':epoch, 'b_period':period, 'host_ldc_telescope':ldc}
#     alles.params_host = {'R_host':r_host, 'M_host':m_host}
#     alles.fill()   
    
#     ind_ecl1, ind_ecl2, ind_out = index_eclipses_smart(time, alles.params['b_epoch'], alles.params['b_period'], alles.params['b_rr'], alles.params['b_rsuma'], alles.params['b_cosi'], alles.params['b_f_s'], alles.params['b_f_c'], 
#                                                         extra_factor=1.1)
#     model_flux = np.ones_like(time)
#     model_flux[ind_ecl1] = alles.generate_model(time[ind_ecl1], inst='telescope', key='flux')
#     inj_flux = model_flux + flux - 1
    
#     if window in ['transit', 'eclipse']:
#         ind_ecl1, ind_ecl2, ind_out = index_eclipses_smart(time, alles.params['b_epoch'], alles.params['b_period'], alles.params['b_rr'], alles.params['b_rsuma'], alles.params['b_cosi'], alles.params['b_f_s'], alles.params['b_f_c'], 
#                                                             extra_factor=10)
#         return time[ind_ecl1], inj_flux[ind_ecl1], flux_err[ind_ecl1]
    
#     elif window is None:
#         return time, inj_flux, flux_err
    


# ::: a fast injection for a simple, circular, and small planet
def inject(time, flux, flux_err, epoch, period, r_companion_earth, r_host=1, m_host=1, ldc=[0.5,0.5], dil=0, return_depth=False, window=None):
    a = ( (G/(4*np.pi**2) * (period*u.d)**2 * (m_host*u.Msun))**(1./3.) ).to(u.AU).value  #in AU 
    r_host_over_a = ((r_host*u.Rsun)/(a*u.AU)).decompose().value #unitless
    r_companion_over_a = ((r_companion_earth*u.Rearth)/(a*u.AU)).decompose().value #unitless
    rr = r_companion_over_a/r_host_over_a
    rsuma = r_companion_over_a + r_host_over_a
    
    ind_ecl1, ind_ecl2, ind_out = index_eclipses_smart(time, epoch, period, rr, rsuma, 0., 0., 0., extra_factor=1.1)
    model_flux = np.ones_like(time)
    if len(ind_ecl1) > 0:
        model_flux[ind_ecl1] = ellc.lc(
                                      t_obs = time[ind_ecl1], 
                                      radius_1 = r_host_over_a, 
                                      radius_2 = r_companion_over_a, 
                                      sbratio = 0., 
                                      incl = 90., 
                                      t_zero = epoch,
                                      period = period,
                                      ldc_1 = ldc,
                                      ld_1 = 'quad',
                                      light_3 = dil / (1. - dil),
                                      verbose = False
                                      )
    inj_flux = model_flux + flux - 1
    
    if return_depth == True:
        depth = 1. - ellc.lc(
                            t_obs = [epoch], 
                            radius_1 = r_host_over_a, 
                            radius_2 = r_companion_over_a, 
                            sbratio = 0., 
                            incl = 90., 
                            t_zero = epoch,
                            period = period,
                            ldc_1 = ldc,
                            ld_1 = 'quad',
                            light_3 = dil / (1. - dil),
                            verbose = False
                            )[0]
        return time, inj_flux, flux_err, depth
    
    else:
        return time, inj_flux, flux_err
    
    # if window in ['transit', 'eclipse']:
    #     ind_ecl1, ind_ecl2, ind_out = index_eclipses_smart(time, epoch, period, rr, rsuma, 0., 0., 0., extra_factor=10)
    #     # inj_flux[ind_out] = np.nan
    #     # return time, inj_flux, flux_err
    #     return time[ind_ecl1], inj_flux[ind_ecl1], flux_err[ind_ecl1]
    # elif window is None:
    #     return time, inj_flux, flux_err

    

###############################################################################
#::: Inject an ellc transit and TLS search on an input lightcurve
###############################################################################
def inject_and_tls_search(time, flux, flux_err, 
                          period_grid, r_companion_earth_grid,
                          dil=0.,
                          known_transits=None, 
                          tls_kwargs=None,
                          wotan_kwargs=None,
                          options=None):
    '''
    Inputs:
    -------
    time : array of flaot
        time stamps of observations
    flux : array of flaot
        normalized flux
    flux_err : array of flaot
        error of normalized flux
    periods : float or array of float
        a period or list of periods for injections
    rplanets : float or array of float
        a planet radius or list of planet radii for injections
        
    known_transits : None or dict
        see allesfitter.detection.tls_search()
        
    tls_kwargs : None or dict
        see allesfitter.detection.tls_search()
        
    wotan_kwargs : None or dict
        see allesfitter.detection.tls_search()
        
    options : None or dict, keywords:
        logfname : str
            file path and name for the log file
            default 
        for the rest see allesfitter.detection.wotan_kwargs()
        
    Summary:
    -------
    - Injects a planet signal via ellc, for a given period and radius (at random epoch)
    - Runs TLS on these injected data and infos
    
    Returns:
    -------
    Nothing, but a list of all TLS results will get saved to a log file
    '''

    #::: format inputs
    period_grid = np.atleast_1d(period_grid)
    r_companion_earth_grid = np.atleast_1d(r_companion_earth_grid)
    
    if tls_kwargs is None: tls_kwargs = {}
    if 'u' not in tls_kwargs: tls_kwargs['u'] = [0.4804, 0.1867] #TLS defaults
    if 'R_star' not in tls_kwargs: tls_kwargs['R_star'] = 1. #TLS defaults
    if 'M_star' not in tls_kwargs: tls_kwargs['M_star'] = 1. #TLS defaults
    if 'SNR_threshold' not in tls_kwargs: tls_kwargs['SNR_threshold'] = 3. #use the lowest SNR to make multiple histograms from SNR 3 to 7
    tls_kwargs['use_threads'] = 1 #no multiprocessing for TLS itself; all multiprocessing is done here
    
    # wotan_kwargs --> allesfitter.detection.tls_search()
    # if wotan_kwargs is None: wotan_kwargs = {}
    # if 'slide_clip' not in wotan_kwargs: wotan_kwargs['slide_clip'] = None  #don't use slide clip (too slow)
    # if 'flatten' not in wotan_kwargs: wotan_kwargs['flatten'] = {} #use a 1 day biweight as default
    # if wotan_kwargs['flatten'] is not None:
    #     if 'method' not in wotan_kwargs['flatten']: wotan_kwargs['flatten']['method'] = 'biweight'
    #     if 'window_length' not in wotan_kwargs['flatten']: wotan_kwargs['flatten']['window_length'] = 1.

    if options is None: options={}
    if 'logfname' not in options: options['logfname'] = 'injection_recovery_test.csv'
    if 'save_plot' not in options: options['save_plot'] = '2' #save TLS plots only
    if 'outdir' not in options: options['outdir'] = 'results'
    if not os.path.exists(options['outdir']): os.makedirs(options['outdir'])
    if 'quiet' not in options: options['quiet'] = True
    if 'multiprocess' not in options: options['multiprocess'] = True
    if 'multiprocess_cores' not in options: options['multiprocess_cores'] = cpu_count()-1
    if 'user_response' not in options: options['user_response'] = None
    

    #::: set up a logfile
    ex = setup_logfile( options )
    
    
    #::: def execution function (for multiprocessing)
    def executer(arg1):
        seed, period, r_companion_earth = arg1
        np.random.seed(seed) #need a fresh seed for every multiprocess
        
        #::: check if it already exists; if not, proceed
        if to_do_or_not_to_do_that_is_the_question(ex, period, r_companion_earth):
           
            #::: draw a random period
            epoch = time[0] + np.random.random()*period
            
            #::: inject transit
            time2, flux2, flux_err2, depth = inject(time, flux, flux_err, epoch, period, r_companion_earth, tls_kwargs['R_star'], tls_kwargs['M_star'], tls_kwargs['u'], dil, return_depth=True, window=None)
            
            #::: pick tight period_min and period_max for TLS, to speed things up
            tls_periods = tls_period_grid(R_star=tls_kwargs['R_star'], M_star=tls_kwargs['M_star'], time_span=time[-1]-time[0])
            ind = np.argmin(np.abs(tls_periods-period))
            tls_kwargs['period_min'] = tls_periods[min(ind+500,len(tls_periods)-1)] #tls_periods is sorted in descending order
            tls_kwargs['period_max'] = tls_periods[max(ind-500,0)] #tls_periods is sorted in descending order

            #::: copy specific options
            options2 = options.copy()
            options2['outdir'] = os.path.join(options['outdir'],'details','P_'+mystr(period)+'_Rp_'+mystr(r_companion_earth))
            options2['inj_period'] = period
            
            #::: run tls with all kwargs
            results_all = tls_search(time2, flux2, flux_err2,
                                     known_transits=known_transits,
                                     tls_kwargs=tls_kwargs,
                                     wotan_kwargs=wotan_kwargs,
                                     options=options2)
            
            #::: write results to file
            if len(results_all)>0:
                for r in results_all:
                    with open(os.path.join(options['outdir'],options['logfname']),'a') as f:
                        f.write(format(period, '.5f')+','+
                                format(r_companion_earth, '.5f')+','+
                                format(depth*1e3, '.5f')+','+ #in ppt
                                format(epoch, '.5f')+','+
                                format(r.period, '.5f')+','+
                                format(r.depth, '.5f')+','+
                                format(r.duration, '.5f')+','+
                                format(r.T0, '.5f')+','+
                                format(r.snr, '.5f')+','+
                                format(r.SDE, '.5f')+','+
                                format(r.FAP, '.5f')+'\n')
            else:
                with open(os.path.join(options['outdir'],options['logfname']),'a') as f:
                    f.write(format(period, '.5f')+','+
                            format(r_companion_earth, '.5f')+','+
                            format(depth*1e3, '.5f')+','+ #in ppt
                            format(epoch, '.5f')+','+
                            'nan'+','+
                            'nan'+','+
                            'nan'+','+
                            'nan'+','+
                            'nan'+','+
                            'nan'+','+
                            'nan'+'\n')
                           
        else:
            pass
        
        
    #::: store all args
    arg_all = []
    seed = 42
    for period, r_companion_earth in itertools.product(period_grid, r_companion_earth_grid): #combining the two for loops
        arg1 = (seed, period, r_companion_earth)
        arg_all.append( arg1 )
        seed += 1
            
        
    #::: cycle through all periods and rplanets
    print('\n', flush=True)
    
    if options['multiprocess'] is False:
        print('Running on a single core')
        for arg1 in tqdm(arg_all):
            executer(arg1)
        
    else:
        print('Running on '+str(options['multiprocess_cores'])+' cores.')
        with closing( Pool(processes=options['multiprocess_cores']) ) as p:
            r = list( tqdm( p.imap(executer, arg_all), total=len(arg_all) ) )
            p.close(); p.join(); p.clear() #needed for pathos, despite closing()
        del p #needed for pathos, despite closing()
                 
        
    #::: finish
    irplot(os.path.join(options['outdir'],options['logfname']), 3, options=options)
    irplot(os.path.join(options['outdir'],options['logfname']), 4, options=options)
    irplot(os.path.join(options['outdir'],options['logfname']), 5, options=options)
    irplot(os.path.join(options['outdir'],options['logfname']), 6, options=options)
    irplot(os.path.join(options['outdir'],options['logfname']), 7, options=options)
    print('Done.')
            


# def get_tls_kwargs_by_tic(tic_id, tls_kwargs=None):
#     u, R_star, R_star_lerr, R_star_uerr, M_star, M_star_lerr, M_star_uerr = catalog_info(TIC_ID=int(tic_id))
#     print('TICv8 info:')
#     print('Quadratic limb darkening u_0, u_1', u[0], u[1])
#     print('Stellar radius', R_star, '+', R_star_lerr, '-', R_star_uerr)
#     print('Stellar mass', M_star, '+', M_star_lerr, '-', M_star_uerr)
#     if tls_kwargs is None: tls_kwargs = {}
#     tls_kwargs['R_star']=float(R_star)
#     tls_kwargs['R_star_min']=R_star-3*R_star_lerr
#     tls_kwargs['R_star_max']=R_star+3*R_star_uerr
#     tls_kwargs['M_star']=float(M_star)
#     tls_kwargs['M_star_min']=M_star-3*M_star_lerr
#     tls_kwargs['M_star_max']=M_star+3*M_star_uerr
#     tls_kwargs['u']=u    
#     return tls_kwargs



###############################################################################
#::: Inject an ellc transit and TLS search using tessio
###############################################################################
def inject_and_tls_search_by_tic(tic_id, 
                                  period_grid, r_companion_earth_grid, 
                                  known_transits=None, 
                                  tls_kwargs=None,
                                  wotan_kwargs=None,
                                  options=None):
    '''
    Inputs:
    -------
    tic_id : str or int
        TIC ID
    periods : float or array of float
        a period or list of periods for injections
    rplanets : float or array of float
        a planet radius or list of planet radii for injections
    logfname : str
        file path and name for the log file
    SNR_threshold : float
        the SNR threshold at which to stop the TLS search
    known_transits : None or dict
        if dict and one transit is already known: 
            known_transits = {'period':[1.3], 'duration':[2.1], 'epoch':[245800.0]}
        if dict and multiple transits are already known: 
            known_transits = {'period':[1.3, 21.0], 'duration':[2.1, 4.1], 'epoch':[245800.0, 245801.0]}
        'period' is the period of the transit
        'duration' must be the total duration, i.e. from first ingress point to last egrees point, in days
        'epoch' is the epoch of the transit
    show_plot : bool
        show a plot in the terminal or not
    save_plot : bool
        save a plot
        
    Summary:
    -------
        - retrieves the SPOC PDC-SAP lightcurve
        - retrieves all TIC catalog information from MAST
        - injects a planet signal via ellc, for a given period and radius (at random epoch)
        - runs TLS on these injected data and infos
    
    Returns:
    -------
        - a list of all TLS results will get saved to a log file
    '''
    
    #::: format inputs
    tic_id = str(int(tic_id))
    period_grid = np.atleast_1d(period_grid)
    r_companion_earth_grid = np.atleast_1d(r_companion_earth_grid)

    #::: load data
    time, flux, flux_err = tessio.get(tic_id, unpack=True)
    if (time is None) or (len(time)<5000): #abort if no reasonable data was found (at least 5000 exposures, aka 7 days of data)
        print('Skipped. No data was found for TIC '+tic_id+'.')
        return None
    
    #::: load tls kwargs
    tls_kwargs = get_tls_kwargs_by_tic(tic_id, tls_kwargs=tls_kwargs)
    if any(np.isnan([tls_kwargs['u'][0], tls_kwargs['u'][1], tls_kwargs['R_star'], tls_kwargs['M_star']])): #abort if TICv8 doesn't hold the stellar info
        print('Skipped. No TICv8 information was found for TIC '+tic_id+'.')
        return None
    
    #::: run
    inject_and_tls_search(time, flux, flux_err, 
                          period_grid, r_companion_earth_grid, 
                          known_transits=known_transits, 
                          tls_kwargs=tls_kwargs,
                          wotan_kwargs=wotan_kwargs,
                          options=options)
            


###############################################################################
#::: diagnostic TLS plots
###############################################################################
def tls_plot(results):
    
    fig, axes = plt.subplots(1,2,figsize=(20,5))
    
    ax = axes[0]
    ax.axvline(results.period, alpha=0.4, lw=3)
    ax.set_xlim(np.min(results.periods), np.max(results.periods))
    for n in range(2, 10):
        ax.axvline(n*results.period, alpha=0.4, lw=1, linestyle="dashed")
        ax.axvline(results.period / n, alpha=0.4, lw=1, linestyle="dashed")
    ax.set_ylabel(r'SDE')
    ax.set_xlabel('Period (days)')
    ax.plot(results.periods, results.power, color='k', lw=0.5)
    ax.set_title('Period '+format(results.period, '.5f') + ', SNR '+format(results.snr, '.1f'))
    ax.set_xlim(0, max(results.periods))
#    plt.show(fig)
#    fig.savefig('plots/'+'Periodogram P = '+str(period)+' days, Rp = '+str(rplanet)+' Rearth'+'.pdf', bbox_inches='tight')
#    plt.close(fig)

    ax = axes[1]
    ax.plot(results.model_folded_phase, results.model_folded_model, color='red')
    ax.scatter(results.folded_phase, results.folded_y, color='b', s=10, alpha=0.5, zorder=2)
    ax.set_xlim(0.4, 0.6)
    ax.set_ylim(0.99,1.01)
    ax.set_xlabel('Phase')
    ax.set_ylabel('Relative flux');
    ax.set_title('Period '+format(results.period, '.5f'))
#    plt.show(fig)
#    fig.savefig('plots/'+'TLS transit P = '+str(period)+' days, Rp = '+str(rplanet)+' Rearth'+'.pdf', bbox_inches='tight')
#    plt.close(fig)
    
    
    
###############################################################################
#::: main
###############################################################################
if __name__ == '__main__':
    pass
    
    ###########################################################################
    #::: Example: injection recovery test with TLS and tessio
    ###########################################################################
    # tic_id = '269701147'
    # periods = np.arange(2,160+eps,2) #in days
    # rplanets = np.arange(0.8,4+eps,0.1) #in R_earth
    # SNR_threshold = 3.
    # known_transits = {'epoch':[2458715.3547, 2458726.0526, 2458743.5534],
    #                   'period':[8.8806, 28.5810, 38.3497],
    #                   'duration':[3.09/24., 4.45/24., 5.52/24.]
    #                   }
    # logfname = 'TIC_269701147.csv'
    
    # inject_and_tls_search_by_tic(tic_id, periods, rplanets, logfname, 
    #                               SNR_threshold=SNR_threshold,
    #                               known_transits=known_transits)
    
    

    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Mar  5 13:02:18 2020

@author: 
Dr. Maximilian N. Günther
MIT Kavli Institute for Astrophysics and Space Research, 
Massachusetts Institute of Technology,
77 Massachusetts Avenue,
Cambridge, MA 02109, 
USA
Email: maxgue@mit.edu
GitHub: https://github.com/MNGuenther
Web: www.mnguenther.com
"""

from __future__ import print_function, division, absolute_import

#::: plotting settings
import seaborn as sns
sns.set(context='paper', style='ticks', palette='deep', font='sans-serif', font_scale=1.5, color_codes=True)
sns.set_style({"xtick.direction": "in","ytick.direction": "in"})
sns.set_context(rc={'lines.markeredgewidth': 1})

#::: modules
import numpy as np
import matplotlib.pyplot as plt
import os, sys
from astropy.constants import G
from astropy import units as u
from pprint import pprint

from . import defaults 




def is_equal(a,b):
    return np.abs(a-b) < 1e-12



def translate_alles_to_ellc(params, settings):
    params2 = params.copy()
    companions_all = list(np.unique(settings['companions_phot'] + settings['companions_rv']))
    inst_all = list(np.unique(settings['inst_phot'] + settings['inst_rv']))
    
    
    #=========================================================================
    #::: inclination, per companion
    #=========================================================================
    for companion in companions_all:
        try:
            params2[companion+'_incl'] = np.arccos( params2[companion+'_cosi'] )/np.pi*180.
        except:
            params2[companion+'_incl'] = None
        
       
    #=========================================================================
    #::: radii, per companion
    #::: R_1/a and R_2/a --> dependent on each companion's orbit
    #=========================================================================
    for companion in companions_all:
        try:
            params2[companion+'_radius_1'] = params2[companion+'_rsuma'] / (1. + params2[companion+'_rr'])
            params2[companion+'_radius_2'] = params2[companion+'_radius_1'] * params2[companion+'_rr']
        except:
            params2[companion+'_radius_1'] = None
            params2[companion+'_radius_2'] = None
            
            
    #=========================================================================
    #::: limb darkening, per instrument
    #=========================================================================
    for inst in inst_all:
        
        #---------------------------------------------------------------------
        #::: host
        #---------------------------------------------------------------------
        if settings['host_ld_law_'+inst] is None:
            params2['host_ldc_'+inst] = None
            
        elif settings['host_ld_law_'+inst] == 'lin':
            params2['host_ldc_'+inst] = params2['host_ldc_q1_'+inst]
            
        elif settings['host_ld_law_'+inst] == 'quad':
            ldc_u1 = 2.*np.sqrt(params2['host_ldc_q1_'+inst]) * params2['host_ldc_q2_'+inst]
            ldc_u2 = np.sqrt(params2['host_ldc_q1_'+inst]) * (1. - 2.*params2['host_ldc_q2_'+inst])
            params2['host_ldc_'+inst] = [ ldc_u1, ldc_u2 ]
            
        elif settings['host_ld_law_'+inst] == 'sing':
            raise ValueError("Sorry, I have not yet implemented the Sing limb darkening law.")
            
        else:
            print(settings['host_ld_law_'+inst] )
            raise ValueError("Currently only 'none', 'lin', 'quad' and 'sing' limb darkening are supported.")
     
    
        #---------------------------------------------------------------------
        #::: companion
        #---------------------------------------------------------------------
        for companion in companions_all:
            
            if settings[companion+'_ld_law_'+inst] is None:
                params2[companion+'_ldc_'+inst] = None
                
            elif settings[companion+'_ld_law_'+inst] == 'lin':
                params2[companion+'_ldc_'+inst] = params2[companion+'_ldc_q1_'+inst]
                
            elif settings[companion+'_ld_law_'+inst] == 'quad':
                ldc_u1 = 2.*np.sqrt(params2[companion+'_ldc_q1_'+inst]) * params2[companion+'_ldc_q2_'+inst]
                ldc_u2 = np.sqrt(params2[companion+'_ldc_q1_'+inst]) * (1. - 2.*params2[companion+'_ldc_q2_'+inst])
                params2[companion+'_ldc_'+inst] = [ ldc_u1, ldc_u2 ]
                
            elif settings[companion+'_ld_law_'+inst] == 'sing':
                raise ValueError("Sorry, I have not yet implemented the Sing limb darkening law.")
                
            else:
                print(settings[companion+'_ld_law_'+inst] )
                raise ValueError("Currently only 'none', 'lin', 'quad' and 'sing' limb darkening are supported.")
            
            
    return params2
    
    
    
def translate(params=None, settings=None, quiet=False, **params_kwargs):
    '''
    A lazy-input translator that calculates whatever is possible from whatever you give it.
        
    epoch : float
        epoch in days
    period : float
        period in days
    R_companion: float
        radius of the companion
        default is 1 Rearth
    M_companion: float
        mass of the companion
        default is 1 Mearth
    R_companion_unit: str
        radius unit of the companion
        default is 'Rearth'
    M_companion_unit: float
        mass unit of the companion
        default is 'Mearth'
    R_host : float
        radius of the star, in Rsun
        default is 1
    M_host: float
        mass of the star, in Msun
        default is 1
    incl : float
        inclination in degrees
    ecc : float
        eccentricity
    omega : float
        argument of periastron in degrees
    ldc : float or list
        limb darkening coefficients
        default is [0.4804, 0.1867]
    dil : float
        dilution, D_0 = 1 - (Ftarget / (Ftarget + Fblend))
        default is 0
    sbratio : float
        surface brightness ratio
        default is 0
        
    '''
    
    
    #==========================================================================
    #::: allow only to either give a params2 dict, or a series of kwargs - not both.
    #==========================================================================
    if params is None:
        params2 = params_kwargs
    elif len(params_kwargs)>0:
        raise ValueError('Give either a params2 dict, or a series of kwargs. Not both.')
    else:
        params2 = params.copy()
        
        
    #==========================================================================
    #::: if no settings are given, use standard settings
    #==========================================================================
    if settings is None:
        settings = defaults.fill_settings() #settings won't get altered, so no copy() needed
        
    
    #==========================================================================
    #::: check R_companion units
    #==========================================================================
    try:
        if 'r_companion_earth' in params2:
            R_companion_with_unit = params2['r_companion_earth']*u.Rearth
        elif 'r_companion_jup' in params2:
            R_companion_with_unit = params2['r_companion_jup']*u.Rjup
        elif 'r_companion_sun' in params2:
            R_companion_with_unit = params2['r_companion_sun']*u.Rsun
    except:
        pass
        
        
    #==========================================================================
    #::: check M_companion units
    #==========================================================================
    try:
        if 'm_companion_earth' in params2:
            M_companion_with_unit = params2['m_companion_earth']*u.Mearth
        elif 'm_companion_jup' in params2:
            M_companion_with_unit = params2['m_companion_jup']*u.Mjup
        elif 'm_companion_sun' in params2:
            M_companion_with_unit = params2['m_companion_sun']*u.Msun
    except:
        pass
    
        
    #==========================================================================
    #::: check for disagreeing inputs
    #==========================================================================
    if 'incl' in params2 and 'cosi' in params2 and not is_equal(params2['incl'], np.rad2deg(np.arccos(params2['cosi']))):
        raise ValueError('Both incl and cosi are given, but are not consistent.')
    
    if 'a' in params2 and 'period' in params2 and 'm_host' in params2:
        if not is_equal( params2['a'], ( (G/(4*np.pi**2) * (params2['period']*u.d)**2 * (params2['m_host']*u.Msun + M_companion_with_unit))**(1./3.) ).to(u.AU).value ):
            raise ValueError('All of a, period and M_host are given, but are not consistent.')
        
    #more TBD
            
        
    #==========================================================================
    #::: display input
    #==========================================================================
    if not quiet: print('\nInput:')
    if not quiet: pprint(params2)
    
    
    #==========================================================================
    #::: check the allowed keys
    #==========================================================================
    # if not quiet: print('\nWarnings:')
    allowed_keys = ['rr',               #allesfitter params2
                    'rsuma',
                    'cosi',
                    'epoch', 
                    'period',
                    'K',
                    'f_c',
                    'f_s',
                    'sbratio',
                    'dil',
                    'ldc',              #in u-space (e.g. Claret 2017)
                    'ldc_transformed',  #in Kipping q-space (eg Kipping 2013)
                    'r_host',           #allesfitter stellar params2
                    'm_host',
                    'incl',             #allesfitter derived params2
                    'a',
                    'ecc',
                    'omega',
                    'r_host_over_a', 
                    'a_over_R_host',
                    'r_companion_over_a',
                    'r_companion_earth',
                    'm_companion_earth',
                    'r_companion_jup',
                    'm_companion_jup',
                    'r_companion_sun',
                    'm_companion_sun']
    for key in list(params2):
        if key not in allowed_keys:
            if not quiet: print('Ignoring the keyword argument "'+key+'", because it was not recognized.')
            del params2[key]
    for key in allowed_keys:
        if key not in list(params2):
            params2[key] = None
            
    
    #==========================================================================
    #::: helper
    #==========================================================================
    def set_(key, value):
        if params2[key] is None:
            params2[key] = np.around( value, decimals=15 ) #round to at most 15 digits to avoid float issues
            
            
    #==========================================================================
    #::: translate
    #==========================================================================
    try: set_('r_companion_earth', params2['rr'] * (params2['r_host']*u.Rsun).to(u.Rearth).value)
    except: pass
    try: set_('r_companion_earth', (params2['r_companion_jup']*u.Rjup).to(u.Rearth).value)
    except: pass
    try: set_('r_companion_earth', (params2['r_companion_sun']*u.Rsun).to(u.Rearth).value)
    except: pass
    
    try: set_('r_companion_jup', params2['rr'] * (params2['r_host']*u.Rsun).to(u.Rjup).value)
    except: pass
    try: set_('r_companion_jup', (params2['r_companion_earth']*u.Rearth).to(u.Rjup).value)
    except: pass
    try: set_('r_companion_jup', (params2['r_companion_sun']*u.Rsun).to(u.Rjup).value)
    except: pass
    
    try: set_('r_companion_sun', params2['rr'] * (params2['r_host']*u.Rsun).to(u.Rsun).value)
    except: pass
    try: set_('r_companion_sun', (params2['r_companion_earth']*u.Rearth).to(u.Rsun).value)
    except: pass
    try: set_('r_companion_sun', (params2['r_companion_jup']*u.Rjup).to(u.Rsun).value)
    except: pass

    try: set_('m_companion_earth', (params2['m_companion_jup']*u.Rjup).to(u.Rearth).value)
    except: pass
    try: set_('m_companion_earth', (params2['m_companion_sun']*u.Rsun).to(u.Rearth).value)
    except: pass
    
    try: set_('m_companion_jup', (params2['m_companion_earth']*u.Rearth).to(u.Rjup).value)
    except: pass
    try: set_('m_companion_jup', (params2['m_companion_sun']*u.Rsun).to(u.Rjup).value)
    except: pass
    
    try: set_('m_companion_sun', (params2['m_companion_earth']*u.Rearth).to(u.Rsun).value)
    except: pass
    try: set_('m_companion_sun', (params2['m_companion_jup']*u.Rjup).to(u.Rsun).value)
    except: pass
        
    try: set_('a', ( (G/(4*np.pi**2) * (params2['period']*u.d)**2 * (params2['m_host']*u.Msun + M_companion_with_unit))**(1./3.) ).to(u.AU).value)  #in AU 
    except: pass

    try: set_('incl', np.rad2deg(np.arccos(params2['cosi']))) #in deg
    except: pass
    try: set_('cosi', np.cos(np.deg2rad(params2['incl'])))
    except: pass
    
    try: set_('ecc', params2['f_c']**2 + params2['f_s']**2) #in deg
    except: pass
    try: set_('omega', np.rad2deg(np.mod( np.arctan2(params2['f_s'], params2['f_c']), 2*np.pi))) #in deg from 0 to 360
    except: pass
    try: set_('f_c', np.sqrt(params2['ecc']) * np.cos(np.deg2rad(params2['omega'])))
    except: pass
    try: set_('f_s', np.sqrt(params2['ecc']) * np.sin(np.deg2rad(params2['omega'])))
    except: pass
    
    try: set_('r_host_over_a', ((params2['r_host']*u.Rsun) / (params2['a']*u.AU)).decompose().value)
    except: pass
    try: set_('r_host_over_a', params['rsuma'] / (1. + params['rr']))
    except: pass
    try: set_('r_host_over_a', 1./params2['a_over_R_host'])
    except: pass


    if params2['a_over_R_host'] is None:
        try: params2['a_over_R_host'] = ((params2['a']*u.AU) / (params2['r_host']*u.Rsun)).decompose().value
        except: pass

    if params2['a_over_R_host'] is None:
        try: params2['a_over_R_host'] = 1./params2['r_host_over_a']
        except: pass

    if params2['r_companion_over_a'] is None:
        try: params2['r_companion_over_a'] = (R_companion_with_unit / (params2['a']*u.AU)).decompose().value
        except: pass

    if params2['r_companion_over_a'] is None:
        try: params2['r_companion_over_a'] = params2['rr'] / params2['a_over_R_host']
        except: pass
    
    if params2['rr'] is None:
        try: params2['rr'] = (R_companion_with_unit / (params2['r_host']*u.Rsun)).decompose().value
        except: pass
    
    if params2['rsuma'] is None:
        try: params2['rsuma'] = ((params2['r_host']*u.Rsun + R_companion_with_unit) / (params2['a']*u.AU)).decompose().value
        except: pass
    
    if params2['rsuma'] is None:
        try: params2['rsuma'] = params2['r_host_over_a'] + params2['r_companion_over_a']
        except: pass
    
    if params2['ldc_transformed'] is None:
        try:
            if settings['ld'] == 'quad':
                q1 = (params2['ldc'][0] + params2['ldc'][1])**2
                q2 = 0.5 * params2['ldc'][0] / (params2['ldc'][0] + params2['ldc'][1])
                params2['ldc_transformed'] = [q1, q2]
        except:
            pass
        
    #more TBD
        
    
    #==========================================================================
    #::: return
    #==========================================================================
    if not quiet: print('\nResults:')
    if not quiet: pprint(params2)
    return params2
    
    

# if __name__ == '__main__':
#     translate(R_companion=1, M_companion=0., cosi=0.1, M_host=0.6, R_host=0.5, period=13.)
#     translate({'rr':0.177, 'a_over_R_host':8.2, 'incl':86, 'omega':346, 'ecc':0.113, 'period':8.360613})



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Jul 10 18:27:01 2020

@author:
Maximilian N. Günther
MIT Kavli Institute for Astrophysics and Space Research, 
Massachusetts Institute of Technology,
77 Massachusetts Avenue,
Cambridge, MA 02109, 
USA
Email: maxgue@mit.edu
Web: www.mnguenther.com
"""

from __future__ import print_function, division, absolute_import

#::: modules
import os, sys
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec
import pandas as pd
from tqdm import tqdm
from glob import glob
from pprint import pprint
from brokenaxes import brokenaxes

#::: my modules
import allesfitter

#::: plotting settings
import seaborn as sns
sns.set(context='paper', style='ticks', palette='deep', font='sans-serif', font_scale=1.5, color_codes=True)
sns.set_style({"xtick.direction": "in","ytick.direction": "in"})
sns.set_context(rc={'lines.markeredgewidth': 1})




def smahtplot(time, flux, subplot_spec=None):
    
    if subplot_spec is None: plt.figure()
    
    ind0 = [0] + list(np.where(np.diff(time)>10)[0]+1) #start indices of data chunks
    ind1 = list(np.where(np.diff(time)>10)[0]) + [len(time)-1] #end indices of data chunks
    xlims = [ (time[i],time[j]) for i,j in zip(ind0,ind1) ]
    
    bax = brokenaxes(xlims=xlims, subplot_spec=subplot_spec)
    bax.plot(time, flux, 'b.')
    
    return bax



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri May  8 12:14:53 2020

@author:
Maximilian N. Günther
MIT Kavli Institute for Astrophysics and Space Research, 
Massachusetts Institute of Technology,
77 Massachusetts Avenue,
Cambridge, MA 02109, 
USA
Email: maxgue@mit.edu
Web: www.mnguenther.com
"""

from __future__ import print_function, division, absolute_import

#::: modules
import os, sys
import numpy as np
import matplotlib.pyplot as plt
import pandas as pd
from tqdm import tqdm
from glob import glob
from pprint import pprint

#::: my modules
import allesfitter

#::: plotting settings
import seaborn as sns
sns.set(context='paper', style='ticks', palette='deep', font='sans-serif', font_scale=1.5, color_codes=True)
sns.set_style({"xtick.direction": "in","ytick.direction": "in"})
sns.set_context(rc={'lines.markeredgewidth': 1})





def mask_ranges(x, x_min, x_max):
    """"
    Crop out values and indices out of an array x for multiple given ranges x_min to x_max.
    
    Input:
    x: array, 
    x_min: lower limits of the ranges
    x_max: upper limits of the ranges
    
    Output:
    
    
    Example:
    x = np.arange(200)    
    x_min = [5, 25, 90]
    x_max = [10, 35, 110]
    """

    mask = np.zeros(len(x), dtype=bool)
    for i in range(len(x_min)): 
        mask = mask | ((x >= x_min[i]) & (x <= x_max[i]))
    ind_mask = np.arange(len(mask))[mask]
    
    return x[mask], ind_mask, mask 



def get_first_epoch(time, epoch, period, width=0):    
    '''
    width : float
        set >0 to include transit egress to mark the first transit
    place the first_epoch at the start of the data to avoid luser mistakes
    '''
    time = np.sort(np.atleast_1d(time))
    start = np.nanmin( time )
    first_epoch = 1.*epoch + width/2. #add width/2 to catch egress
    if start<=first_epoch: first_epoch -= np.floor((first_epoch-start)/period) * period
    else: first_epoch += np.ceil((start-first_epoch)/period) * period
    return first_epoch - width/2.  #subtract width/2 to get midpoint again
    
    

def get_epoch_occ(epoch, period, f_s, f_c):
    ecc = f_s**2 + f_c**2
    ecosw = f_c * np.sqrt(ecc)
    epoch_occ = epoch + period/2. * (1. + 4./np.pi * ecosw)
    return epoch_occ



def get_Rhost_over_a(rr, rsuma):
    Rhost_over_a = rsuma / (1. + rr)
    return Rhost_over_a
    


def get_ecc_esinw_ecosw(f_s, f_c):
    ecc = f_s**2 + f_c**2
    esinw = f_s * np.sqrt(ecc)
    ecosw = f_c * np.sqrt(ecc)
    return ecc, esinw, ecosw
    
    

def impact_parameters_smart(rr, rsuma, cosi, f_s, f_c):
    #--------------------------------------------------------------------------
    #::: inputs
    #--------------------------------------------------------------------------
    Rhost_over_a = get_Rhost_over_a(rr, rsuma)
    ecc, esinw, ecosw = get_ecc_esinw_ecosw(f_s, f_c)
    
    
    #--------------------------------------------------------------------------
    #::: impact parameters
    #--------------------------------------------------------------------------
    eccentricity_correction_b_1 = ( (1. - ecc**2) / ( 1. + esinw ) )
    b_1 = (1./Rhost_over_a) * cosi * eccentricity_correction_b_1
    
    eccentricity_correction_b_2 = ( (1. - ecc**2) / ( 1. - esinw ) )
    b_2 = (1./Rhost_over_a) * cosi * eccentricity_correction_b_2
    
    
    #--------------------------------------------------------------------------
    #::: return
    #--------------------------------------------------------------------------
    return b_1, b_2

    

def eclipse_width_smart(period, rr, rsuma, cosi, f_s, f_c):
    #--------------------------------------------------------------------------
    #::: inputs
    #--------------------------------------------------------------------------
    Rhost_over_a = get_Rhost_over_a(rr, rsuma)
    ecc, esinw, ecosw = get_ecc_esinw_ecosw(f_s, f_c)
    
    
    #--------------------------------------------------------------------------
    #::: impact parameters 
    #--------------------------------------------------------------------------
    b_1, b_2 = impact_parameters_smart(rr, rsuma, cosi, f_s, f_c)
    
    
    #--------------------------------------------------------------------------
    #::: widths
    #--------------------------------------------------------------------------
    eccentricity_correction_width_1 = ( np.sqrt(1. - ecc**2) / ( 1. + esinw ) )
    width_1 = period/np.pi  \
            * np.arcsin( Rhost_over_a \
                         * np.sqrt( (1. + rr)**2 - b_1**2  )\
                         / np.sin(np.arccos(cosi)) ) \
            * eccentricity_correction_width_1
    
    width_2 = width_1 * (1. + esinw) / (1. - esinw)
    
    
    #--------------------------------------------------------------------------
    #::: return
    #--------------------------------------------------------------------------
    return width_1, width_2
    
    

def index_eclipses_smart(time, epoch, period, rr, rsuma, cosi, f_s, f_c, extra_factor=1.):
    '''
    
    Parameters
    ----------
    time : array
        must be sorted
    epoch : float
        must be first_epoch in the data set
    period : float
        DESCRIPTION.
    rr : float
        DESCRIPTION.
    rsuma : float
        DESCRIPTION.
    cosi : float
        DESCRIPTION.
    f_s : float
        DESCRIPTION.
    f_c : float
        DESCRIPTION.
    extra_factor : float, optional
        DESCRIPTION. The default is 1..

    Returns
    -------
    ind_ecl1 : array
        DESCRIPTION.
    ind_ecl2 : array
        DESCRIPTION.
    ind_out : array
        DESCRIPTION.

    '''
    #--------------------------------------------------------------------------
    #::: inputs
    #--------------------------------------------------------------------------
    ecc, esinw, ecosw = get_ecc_esinw_ecosw(f_s, f_c)
    
    
    #--------------------------------------------------------------------------
    #::: widths
    #--------------------------------------------------------------------------
    width_1, width_2 = eclipse_width_smart(period, rr, rsuma, cosi, f_s, f_c)
    
    
    #--------------------------------------------------------------------------
    #::: timing
    #--------------------------------------------------------------------------
    # time = np.sort(time)
    # epoch = get_first_epoch(time, epoch, period, width=2*width_1)
    epoch_occ = epoch + period/2. * (1. + 4./np.pi * ecosw)
    
    N = int( 1. * ( time[-1] - epoch ) / period ) + 1
        
    tmid_ecl1 = np.array( [ epoch + i * period for i in range(N) ] )
    tmid_ecl2 = np.array( [ epoch_occ + (i-1) * period for i in range(N+1) ] ) #(o-1) in case an occ happens before the first epoch
    
    _, ind_ecl1, mask_ecl1 = mask_ranges( time, tmid_ecl1 - width_1/2.*extra_factor, tmid_ecl1 + width_1/2.*extra_factor )           
    _, ind_ecl2, mask_ecl2 = mask_ranges( time, tmid_ecl2 - width_2/2.*extra_factor, tmid_ecl2 + width_2/2.*extra_factor )
        
    ind_out = np.arange( len(time) )[ ~(mask_ecl1 | mask_ecl2) ]
    
    
    #--------------------------------------------------------------------------
    #::: return
    #--------------------------------------------------------------------------
    return ind_ecl1, ind_ecl2, ind_out



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Oct  5 00:41:29 2018

@author:
Maximilian N. Günther
MIT Kavli Institute for Astrophysics and Space Research, 
Massachusetts Institute of Technology,
77 Massachusetts Avenue,
Cambridge, MA 02109, 
USA
Email: maxgue@mit.edu
Web: www.mnguenther.com
"""

from __future__ import print_function, division, absolute_import

#::: plotting settings
import seaborn as sns
sns.set(context='paper', style='ticks', palette='deep', font='sans-serif', font_scale=1.5, color_codes=True)
sns.set_style({"xtick.direction": "in","ytick.direction": "in"})
sns.set_context(rc={'lines.markeredgewidth': 1})

#::: modules
import numpy as np
import ellc
from scipy.optimize import minimize
from scipy.interpolate import UnivariateSpline
import numpy.polynomial.polynomial as poly
import warnings
warnings.filterwarnings('ignore', category=np.VisibleDeprecationWarning) 
warnings.filterwarnings('ignore', category=np.RankWarning) 
warnings.filterwarnings('ignore', category=RuntimeWarning) 
try:
    import celerite
    from celerite import terms
except ImportError:
    warnings.warn("Cannot import package 'celerite', thus 'hybrid_GP' baseline models will not be supported.")

#allesfitter modules
from . import config
from .limb_darkening import LDC3
from .flares.aflare import aflare1
# from .exoworlds_rdx.lightcurves.lightcurve_tools import calc_phase


'''
README

for dilution:
    our definition of dilution is D_0 = 1 - F_source / (F_source + F_blend)
    ellc's original definition of third light is light_3 = F_blend / F_source
    therefore we can relate them as light_3 = D_0 / (1 - D_0)
    
    on another note, the TESS SPOC lightcurve parameter CROWDSAP = F_source / (F_source + F_blend) d
    hence D_0 = 1 - CROWDSAP
'''






###############################################################################
#::: possible functions
###############################################################################  
GPs = ['sample_GP_real', 'sample_GP_complex', 'sample_GP_Matern32', 'sample_GP_SHO']
FCTs = ['none', 'offset', 'linear', 'hybrid_offset', 'hybrid_poly_0', 'hybrid_poly_1', 'hybrid_poly_2', 'hybrid_poly_3', 'hybrid_poly_4', 'hybrid_poly_5', 'hybrid_poly_6', 'hybrid_spline', 'sample_offset', 'sample_linear']



###############################################################################
#::: divider that catches 'None' types
###############################################################################  
def divide(a,b):
    if a is not None:
        return 1.*a/b
    else:
        return None
    
    

###############################################################################
#::: convert input params into ellc params
###############################################################################  
def update_params(theta):
    
    params = config.BASEMENT.params.copy()
    
    
    #=========================================================================
    #::: first, sync over from theta
    #=========================================================================
    for i, key in enumerate(config.BASEMENT.fitkeys):
        params[key] = theta[i]   
    
    
    #=========================================================================
    #::: second, deal with coupled params before updates
    #=========================================================================
    for i, key in enumerate(config.BASEMENT.allkeys):
        if isinstance(config.BASEMENT.coupled_with[i], str) and (len(config.BASEMENT.coupled_with[i])>0):
            params[key] = params[config.BASEMENT.coupled_with[i]]
    
    
    #=========================================================================
    #::: inclination, per companion
    #=========================================================================
    for companion in config.BASEMENT.settings['companions_all']:
        try:
            params[companion+'_incl'] = np.arccos( params[companion+'_cosi'] )/np.pi*180.
        except:
            params[companion+'_incl'] = None
        
       
    #=========================================================================
    #::: radii, per companion
    #::: R_1/a and R_2/a --> dependent on each companion's orbit
    #=========================================================================
    for companion in config.BASEMENT.settings['companions_all']:
        try:
            params[companion+'_radius_1'] = params[companion+'_rsuma'] / (1. + params[companion+'_rr'])
            params[companion+'_radius_2'] = params[companion+'_radius_1'] * params[companion+'_rr']
        except:
            params[companion+'_radius_1'] = None
            params[companion+'_radius_2'] = None
                
                
    #=========================================================================
    #::: limb darkening, per instrument
    #=========================================================================
    for inst in config.BASEMENT.settings['inst_all']:
        
        #---------------------------------------------------------------------
        #::: host
        #---------------------------------------------------------------------
        if config.BASEMENT.settings['host_ld_law_'+inst] is None:
            params['host_ldc_'+inst] = None
            
        elif config.BASEMENT.settings['host_ld_law_'+inst] == 'lin':
            params['host_ldc_'+inst] = params['host_ldc_q1_'+inst]
            
        elif config.BASEMENT.settings['host_ld_law_'+inst] == 'quad':
            ldc_u1 = 2.*np.sqrt(params['host_ldc_q1_'+inst]) * params['host_ldc_q2_'+inst]
            ldc_u2 = np.sqrt(params['host_ldc_q1_'+inst]) * (1. - 2.*params['host_ldc_q2_'+inst])
            params['host_ldc_'+inst] = [ ldc_u1, ldc_u2 ]
            
        elif config.BASEMENT.settings['host_ld_law_'+inst] == 'sing':
            params['host_ldc_'+inst] = LDC3.forward([params['host_ldc_q1_'+inst], params['host_ldc_q2_'+inst], params['host_ldc_q3_'+inst]])

        else:
            raise ValueError("Currently only 'none', 'lin', 'quad' and 'sing' limb darkening are supported, but "+config.BASEMENT.settings['host_ld_law_'+inst]+" was input.")
    
    
        #---------------------------------------------------------------------
        #::: companion
        #---------------------------------------------------------------------
        for companion in config.BASEMENT.settings['companions_all']:
            
            if config.BASEMENT.settings[companion+'_ld_law_'+inst] is None:
                params[companion+'_ldc_'+inst] = None
                
            elif config.BASEMENT.settings[companion+'_ld_law_'+inst] == 'lin':
                params[companion+'_ldc_'+inst] = params[companion+'_ldc_q1_'+inst]
                
            elif config.BASEMENT.settings[companion+'_ld_law_'+inst] == 'quad':
                ldc_u1 = 2.*np.sqrt(params[companion+'_ldc_q1_'+inst]) * params[companion+'_ldc_q2_'+inst]
                ldc_u2 = np.sqrt(params[companion+'_ldc_q1_'+inst]) * (1. - 2.*params[companion+'_ldc_q2_'+inst])
                params[companion+'_ldc_'+inst] = [ ldc_u1, ldc_u2 ]
                
            elif config.BASEMENT.settings[companion+'_ld_law_'+inst] == 'sing':
                params[companion+'_ldc_'+inst] = LDC3.forward([params[companion+'_ldc_q1_'+inst], params[companion+'_ldc_q2_'+inst], params[companion+'_ldc_q3_'+inst]])
                
            else:
                raise ValueError("Currently only 'none', 'lin', 'quad' and 'sing' limb darkening are supported, but "+config.BASEMENT.settings[companion+'_ld_law_'+inst]+" was input.")
    
    
    #=========================================================================
    #::: photometric errors, per instrument
    #=========================================================================
    for inst in config.BASEMENT.settings['inst_phot']:
        key='flux'
        params['err_'+key+'_'+inst] = np.exp( params['ln_err_'+key+'_'+inst] )
        
        
    #=========================================================================
    #::: RV jitter, per instrument
    #=========================================================================
    for inst in config.BASEMENT.settings['inst_rv']:
        key='rv'
        params['jitter_'+key+'_'+inst] = np.exp( params['ln_jitter_'+key+'_'+inst] )
        
        
    #=========================================================================
    #::: semi-major axes, per companion
    #=========================================================================
    for companion in config.BASEMENT.settings['companions_all']:
        params[companion+'_ecc'] = params[companion+'_f_s']**2 + params[companion+'_f_c']**2
        try:
            a_1 = 0.019771142 * params[companion+'_K'] * params[companion+'_period'] * np.sqrt(1. - params[companion+'_ecc']**2)/np.sin(params[companion+'_incl']*np.pi/180.)
            params[companion+'_a'] = (1.+1./params[companion+'_q'])*a_1
        except:
            params[companion+'_a'] = None
        if params[companion+'_a'] == 0.:
            params[companion+'_a'] = None
               
                
    #=========================================================================
    #::: stellar spots, per companion and instrument
    #=========================================================================
    for companion in config.BASEMENT.settings['companions_all']:
        for inst in config.BASEMENT.settings['inst_all']:
            
            #---------------------------------------------------------------------
            #::: host spots
            #---------------------------------------------------------------------
            if config.BASEMENT.settings['host_N_spots_'+inst] > 0:
                params['host_spots_'+inst] = [
                                     [params['host_spot_'+str(i)+'_long_'+inst] for i in range(1,config.BASEMENT.settings['host_N_spots_'+inst]+1) ],
                                     [params['host_spot_'+str(i)+'_lat_'+inst] for i in range(1,config.BASEMENT.settings['host_N_spots_'+inst]+1) ],
                                     [params['host_spot_'+str(i)+'_size_'+inst] for i in range(1,config.BASEMENT.settings['host_N_spots_'+inst]+1) ],
                                     [params['host_spot_'+str(i)+'_brightness_'+inst] for i in range(1,config.BASEMENT.settings['host_N_spots_'+inst]+1) ]
                                    ]
        
            #---------------------------------------------------------------------
            #::: companion spots
            #---------------------------------------------------------------------
            if config.BASEMENT.settings[companion+'_N_spots_'+inst] > 0:
                params[companion+'_spots_'+inst] = [
                                     [params[companion+'_spot_'+str(i)+'_long_'+inst] for i in range(1,config.BASEMENT.settings[companion+'_N_spots_'+inst]+1) ],
                                     [params[companion+'_spot_'+str(i)+'_lat_'+inst] for i in range(1,config.BASEMENT.settings[companion+'_N_spots_'+inst]+1) ],
                                     [params[companion+'_spot_'+str(i)+'_size_'+inst] for i in range(1,config.BASEMENT.settings[companion+'_N_spots_'+inst]+1) ],
                                     [params[companion+'_spot_'+str(i)+'_brightness_'+inst] for i in range(1,config.BASEMENT.settings[companion+'_N_spots_'+inst]+1) ]
                                    ]
        
        
    #=========================================================================
    #::: host stellar density, per companion
    #::: (in cgs units)
    #::: via Seager & Mallen-Ornelas 2003 (see also Winn 2010, Eq. 30)
    #::: this needs rr^3 := (R_companion/R_star)^3 --> 0 
    #::: hence we here demand rr < 0.215443469, allowing a 1% erroneous contribution by the planet density).
    #=========================================================================
    for companion in config.BASEMENT.settings['companions_phot']:
        if (params[companion+'_rr'] is not None) and (params[companion+'_rr'] > 0) and (params[companion+'_rr'] < 0.215443469):
            params[companion+'_host_density'] = 3. * np.pi * (1./params[companion+'_radius_1'])**3. / (params[companion+'_period']*86400.)**2 / 6.67408e-8 #in cgs
        else:
            params[companion+'_host_density'] = None
            
        
    #=========================================================================
    #::: lastly, deal with coupled params again after updates
    #=========================================================================
    for i, key in enumerate(config.BASEMENT.allkeys):
        if isinstance(config.BASEMENT.coupled_with[i], str) and (len(config.BASEMENT.coupled_with[i])>0):
            params[key] = params[config.BASEMENT.coupled_with[i]]
            
            
    return params




###############################################################################
#::: flux fct
###############################################################################
    
#==============================================================================
#::: flux fct: main
#==============================================================================
def flux_fct(params, inst, companion, xx=None, settings=None):
    '''
    ! params must be updated via update_params() before calling this function !
    '''
    
    if settings is None: 
        settings = config.BASEMENT.settings
    
    if settings['fit_ttvs']==False:
        return flux_fct_full(params, inst, companion, xx=xx, settings=settings)
    else:
        return flux_fct_piecewise(params, inst, companion, xx=xx, settings=settings)



#==============================================================================
#::: flux fct: full curve (no TTVs)
#==============================================================================
def flux_fct_full(params, inst, companion, xx=None, settings=None):
    '''
    ! params must be updated via update_params() before calling this function !
    '''
    
    #-------------------------------------------------------------------------- 
    #::: defaults
    #-------------------------------------------------------------------------- 
    if settings is None: 
        settings = config.BASEMENT.settings
        
    if xx is None: 
        xx = config.BASEMENT.data[inst]['time']
        t_exp = settings['t_exp_'+inst]
        n_int = settings['t_exp_n_int_'+inst]
    else:
        t_exp = None
        n_int = None
    
    
    #-------------------------------------------------------------------------- 
    #::: flux sub-fct: ellc lightcurve 
    # for eclipses, transits, occultations, 
    # spots, rotation,
    # and physical phase curve models; but no phase shift and no thermal vs reflected
    #-------------------------------------------------------------------------- 
    model_flux, model_flux1, model_flux2 = flux_subfct_ellc(params, inst, companion, xx=xx, settings=settings, t_exp=t_exp, n_int=n_int, return_fluxes=True)
        
        
    #-------------------------------------------------------------------------- 
    #::: flux sub-fct: ellc hacked phase curves
    # allowing phase shifts and thermal vs reflected
    #-------------------------------------------------------------------------- 
    '''
    if (companion+'_thermal_emission_amplitude_'+inst in params) and (params[companion+'_thermal_emission_amplitude_'+inst]>0):
        model_flux += calc_thermal_curve(params, inst, companion, xx, t_exp, n_int)
    '''
    
    
    #-------------------------------------------------------------------------- 
    #::: flux sub-fct: sinusoidal phase curves
    # allowing phase shifts and thermal vs reflected
    #-------------------------------------------------------------------------- 
    if settings['phase_curve_style'] in ['sine_series', 'sine_physical']:
        model_flux += flux_subfct_sinusoidal_phase_curves(params, inst, companion, model_flux2, xx=xx, settings=settings) - 1.
    
        
    #-------------------------------------------------------------------------- 
    #::: flux sub-fct: flare models
    #-------------------------------------------------------------------------- 
    if settings['N_flares'] > 0:
        model_flux += flux_subfct_flares(params, inst, companion, xx=xx, settings=settings) - 1.
    
    
    #-------------------------------------------------------------------------- 
    #::: return
    #-------------------------------------------------------------------------- 
    return model_flux



#==============================================================================
#::: flux sub-fct: ellc lightcurve
# for transits, occultations, eclipses, 
# and physical phase curve models (no phase shift; no thermal vs reflected)
#==============================================================================
def flux_subfct_ellc(params, inst, companion, xx=None, settings=None, t_exp=None, n_int=None, return_fluxes=False):
    '''
    ! params must be updated via update_params() before calling this function !
    '''
    
    #-------------------------------------------------------------------------- 
    #::: defaults
    #-------------------------------------------------------------------------- 
    if settings is None: 
        settings = config.BASEMENT.settings
        
    if xx is None: 
        xx = config.BASEMENT.data[inst]['time']
        t_exp = settings['t_exp_'+inst]
        n_int = settings['t_exp_n_int_'+inst]
    # else:
    #     t_exp = None
    #     n_int = None
    

    #-------------------------------------------------------------------------- 
    #::: index transits and occultations
    #-------------------------------------------------------------------------- 
    # ind_ecl1, ind_ecl2, ind_out = index_eclipses_smart(xx, params[companion+'_epoch'], params[companion+'_period'], params[companion+'_rr'], params[companion+'_rsuma'], params[companion+'_cosi'], params[companion+'_f_s'], params[companion+'_f_c'], extra_factor=1.5)
    #TODO: possible future speed improvement
        
        
    #-------------------------------------------------------------------------- 
    #::: if: planet and EB lightcurve model
    #-------------------------------------------------------------------------- 
    if (params[companion+'_rr'] is not None) and (params[companion+'_rr'] > 0):
        model_flux1, model_flux2 = ellc.fluxes(
                                    t_obs =       xx, 
                                    radius_1 =    params[companion+'_radius_1'], 
                                    radius_2 =    params[companion+'_radius_2'], 
                                    sbratio =     params[companion+'_sbratio_'+inst], 
                                    incl =        params[companion+'_incl'], 
                                    # light_3 =     params['dil_'+inst] / (1.-params['dil_'+inst]), #fluxes does not take light_3
                                    t_zero =      params[companion+'_epoch'],
                                    period =      params[companion+'_period'],
                                    a =           params[companion+'_a'],
                                    q =           params[companion+'_q'],
                                    f_c =         params[companion+'_f_c'],
                                    f_s =         params[companion+'_f_s'],
                                    ldc_1 =       params['host_ldc_'+inst],
                                    ldc_2 =       params[companion+'_ldc_'+inst],
                                    gdc_1 =       params['host_gdc_'+inst],
                                    gdc_2 =       params[companion+'_gdc_'+inst],
                                    didt =        params['didt_'+inst], 
                                    domdt =       params['domdt_'+inst], 
                                    rotfac_1 =    params['host_rotfac_'+inst], 
                                    rotfac_2 =    params[companion+'_rotfac_'+inst], 
                                    hf_1 =        params['host_hf_'+inst], #1.5, 
                                    hf_2 =        params[companion+'_hf_'+inst], #1.5,
                                    bfac_1 =      params['host_bfac_'+inst],
                                    bfac_2 =      params[companion+'_bfac_'+inst], 
                                    heat_1 =      divide(params['host_heat_'+inst],2.),
                                    heat_2 =      divide(params[companion+'_heat_'+inst],2.),
                                    lambda_1 =    params['host_lambda_'+inst], 
                                    lambda_2 =    params[companion+'_lambda_'+inst], 
                                    vsini_1 =     params['host_vsini'],
                                    vsini_2 =     params[companion+'_vsini'], 
                                    t_exp =       t_exp,
                                    n_int =       n_int,
                                    grid_1 =      settings['host_grid_'+inst],
                                    grid_2 =      settings[companion+'_grid_'+inst],
                                    ld_1 =        settings['host_ld_law_'+inst],
                                    ld_2 =        settings[companion+'_ld_law_'+inst],
                                    shape_1 =     settings['host_shape_'+inst],
                                    shape_2 =     settings[companion+'_shape_'+inst],
                                    spots_1 =     params['host_spots_'+inst], 
                                    spots_2 =     params[companion+'_spots_'+inst], 
                                    exact_grav =  settings['exact_grav'],
                                    verbose =     False
                                    )
  
        #::: combine the host and companion fluxes, and account for dilution
        model_flux = 1. + ( (model_flux1+model_flux2-1.) * (1.-params['dil_'+inst]) )  


    #-------------------------------------------------------------------------- 
    #::: else: constant 1
    #-------------------------------------------------------------------------- 
    else:
        model_flux1 = np.ones_like(xx)
        model_flux2 = np.zeros_like(xx)
        model_flux = np.ones_like(xx)

        
    #-------------------------------------------------------------------------- 
    #::: return
    #-------------------------------------------------------------------------- 
    if not return_fluxes:
        return model_flux
    else:
        return model_flux, model_flux1, model_flux2
    
    

#==============================================================================
#::: flux sub-fct: sinusoidal phase curves, allowing phase shifts
#==============================================================================
def flux_subfct_sinusoidal_phase_curves(params, inst, companion, model_flux2, xx=None, settings=None):    
    '''
    ! params must be updated via update_params() before calling this function !
    '''
    
    #-------------------------------------------------------------------------- 
    #::: defaults
    #-------------------------------------------------------------------------- 
    if settings is None:
        settings = config.BASEMENT.settings
        
    if xx is None:
        xx = config.BASEMENT.data[inst]['time']
        
    model_flux = np.ones_like(xx)
        
        
    #-------------------------------------------------------------------------- 
    #::: derive normalized curve for fractional loss of "atmospheric light" 
    # during occultation / secondary eclipse 
    #-------------------------------------------------------------------------- 
    if all(model_flux2==0.): flux2_norm = 1.
    else: flux2_norm = model_flux2/np.nanmax(model_flux2)
    
    
    #-------------------------------------------------------------------------- 
    #::: sine/cosine phase curve model
    # A1 (beaming)
    # B1 (atmospheric), can be split in thermal and reflected
    # B2 (ellipsoidal)
    # B3 (ellipsoidal 2nd order)
    #-------------------------------------------------------------------------- 
    
    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    #::: the standard sine/cosine definition, e.g. used by Shporer and Wong
    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    if settings['phase_curve_style'] == 'sine_series':
        
        if (params[companion+'_phase_curve_A1_'+inst] is not None): #A1 (beaming)
            model_flux += (1.-params['dil_'+inst]) * 1e-3*params[companion+'_phase_curve_A1_'+inst] * np.sin(2.*np.pi/params[companion+'_period'] * (xx - params[companion+'_epoch']))
            
        if (params[companion+'_phase_curve_B1_'+inst] is not None): #B1 (atmospheric)
            model_flux += (1.-params['dil_'+inst]) * 1e-3*params[companion+'_phase_curve_B1_'+inst] * (flux2_norm * (np.cos(2.*np.pi/params[companion+'_period'] * (xx - params[companion+'_epoch'] + params[companion+'_phase_curve_B1_shift_'+inst])) - 1) + 1 )

        if (params[companion+'_phase_curve_B1t_'+inst] is not None): #B1 (atmospheric) thermal part
            model_flux += (1.-params['dil_'+inst]) * 1e-3*params[companion+'_phase_curve_B1t_'+inst] * (flux2_norm * (np.cos(2.*np.pi/params[companion+'_period'] * (xx - params[companion+'_epoch'] + params[companion+'_phase_curve_B1t_shift_'+inst])) - 1) + 1 )

        if (params[companion+'_phase_curve_B1r_'+inst] is not None): #B1 (atmospheric) reflected part
            model_flux += (1.-params['dil_'+inst]) * 1e-3*params[companion+'_phase_curve_B1r_'+inst] * (flux2_norm * (np.cos(2.*np.pi/params[companion+'_period'] * (xx - params[companion+'_epoch'] + params[companion+'_phase_curve_B1r_shift_'+inst])) - 1) + 1 )

        if (params[companion+'_phase_curve_B2_'+inst] is not None): #B2 (ellipsoidal)
            model_flux += (1.-params['dil_'+inst]) * 1e-3*params[companion+'_phase_curve_B2_'+inst] * np.cos(2. * 2.*np.pi/params[companion+'_period'] * (xx - params[companion+'_epoch']))

        if (params[companion+'_phase_curve_B3_'+inst] is not None): #B3 (ellipsoidal)
            model_flux += (1.-params['dil_'+inst]) * 1e-3*params[companion+'_phase_curve_B3_'+inst] * np.cos(3. * 2.*np.pi/params[companion+'_period'] * (xx - params[companion+'_epoch']))

        #::: for debugging:
            
        # if (params[companion+'_phase_curve_A1_'+inst] is not None): #A1 (beaming)
        #     flux_A1 = (1.-params['dil_'+inst]) * 1e-3*params[companion+'_phase_curve_A1_'+inst] * np.sin(2.*np.pi/params[companion+'_period'] * (xx - params[companion+'_epoch']))
            
        # if (params[companion+'_phase_curve_B1_'+inst] is not None): #B1 (atmospheric)
        #     flux_B1 = (1.-params['dil_'+inst]) * 1e-3*params[companion+'_phase_curve_B1_'+inst] * (flux2_norm * (np.cos(2.*np.pi/params[companion+'_period'] * (xx - params[companion+'_epoch'] + params[companion+'_phase_curve_B1_shift_'+inst])) - 1) + 1 )

        # if (params[companion+'_phase_curve_B2_'+inst] is not None): #B2 (ellipsoidal)
        #     flux_B2 = (1.-params['dil_'+inst]) * 1e-3*params[companion+'_phase_curve_B2_'+inst] * np.cos(2. * 2.*np.pi/params[companion+'_period'] * (xx - params[companion+'_epoch']))

        # import matplotlib.pyplot as plt
        
        # plt.figure()
        # plt.plot(xx, flux2_norm, label='flux2_norm')
        # plt.legend()
        
        # plt.figure()
        # plt.plot(xx, flux_A1, label='A1')
        # plt.plot(xx, flux_B1, label='B1')
        # plt.plot(xx, flux_B2, label='B2')
        # plt.ylim([0.999-1,1.001-1])
        # plt.legend()
        

    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    #::: the additive sine/cosine definition, e.g. used by Daylan
    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    elif settings['phase_curve_style'] == 'sine_physical':
        
        if (params[companion+'_phase_curve_beaming_'+inst] is not None): #A1, rescaled
            model_flux += (1.-params['dil_'+inst]) * 1e-3*params[companion+'_phase_curve_beaming_'+inst] * np.sin(2.*np.pi/params[companion+'_period'] * (xx - params[companion+'_epoch']))
            
        if (params[companion+'_phase_curve_atmospheric_'+inst] is not None): #B1, rescaled
            model_flux += (1.-params['dil_'+inst]) * 1e-3*params[companion+'_phase_curve_atmospheric_'+inst] * flux2_norm * 0.5 * (1. - np.cos(2.*np.pi/params[companion+'_period'] * (xx - params[companion+'_epoch'] + params[companion+'_phase_curve_atmospheric_shift_'+inst])))
    
        if (params[companion+'_phase_curve_atmospheric_thermal_'+inst] is not None): #B1 thermal part, rescaled
            model_flux += (1.-params['dil_'+inst]) * 1e-3*params[companion+'_phase_curve_atmospheric_thermal_'+inst] * flux2_norm * 0.5 * (1. - np.cos(2.*np.pi/params[companion+'_period'] * (xx - params[companion+'_epoch'] + params[companion+'_phase_curve_atmospheric_thermal_shift_'+inst])))
    
        if (params[companion+'_phase_curve_atmospheric_reflected_'+inst] is not None): #B1 reflected part, rescaled
            model_flux += (1.-params['dil_'+inst]) * 1e-3*params[companion+'_phase_curve_atmospheric_reflected_'+inst] * flux2_norm * 0.5 * (1. - np.cos(2.*np.pi/params[companion+'_period'] * (xx - params[companion+'_epoch'] + params[companion+'_phase_curve_atmospheric_reflected_shift_'+inst])))
    
        if (params[companion+'_phase_curve_ellipsoidal_'+inst] is not None): #B2, rescaled
            model_flux += (1.-params['dil_'+inst]) * 1e-3*params[companion+'_phase_curve_ellipsoidal_'+inst] * 0.5 * (1. - np.cos(2. * 2.*np.pi/params[companion+'_period'] * (xx - params[companion+'_epoch'])))
   
        if (params[companion+'_phase_curve_ellipsoidal_2nd_'+inst] is not None): #B3, rescaled
            model_flux += (1.-params['dil_'+inst]) * 1e-3*params[companion+'_phase_curve_ellipsoidal_2nd_'+inst] * 0.5 * (1. - np.cos(3. * 2.*np.pi/params[companion+'_period'] * (xx - params[companion+'_epoch'])))
   
        
    # err
    #-------------------------------------------------------------------------- 
    #::: return
    #-------------------------------------------------------------------------- 
    return model_flux
    


#==============================================================================
#::: flux sub-fct: flare models
#==============================================================================
def flux_subfct_flares(params, inst, companion, xx=None, settings=None, return_fluxes=False):
    
    #-------------------------------------------------------------------------- 
    #::: defaults
    #-------------------------------------------------------------------------- 
    if settings is None:
        settings = config.BASEMENT.settings
        
    if xx is None:
        xx = config.BASEMENT.data[inst]['time']
        
    model_flux = np.ones_like(xx)
    
    
    #-------------------------------------------------------------------------- 
    #::: flares
    #-------------------------------------------------------------------------- 
    if settings['N_flares'] > 0:
        for i in range(1,settings['N_flares']+1):
            model_flux += (1.-params['dil_'+inst]) * aflare1(xx, params['flare_tpeak_'+str(i)], params['flare_fwhm_'+str(i)], params['flare_ampl_'+str(i)], upsample=True, uptime=10)
    
    
    #-------------------------------------------------------------------------- 
    #::: return
    #-------------------------------------------------------------------------- 
    return model_flux
    


#==============================================================================
#::: flux sub-fct: ellc lightcurves piecewise (for TTVs; no phase curve)
#==============================================================================
def flux_fct_piecewise(params, inst, companion, xx=None, settings=None):
    '''
    ! params must be updated via update_params() before calling this function !
    '''
    
    #-------------------------------------------------------------------------- 
    #::: defaults
    #-------------------------------------------------------------------------- 
    if settings is None:
        settings = config.BASEMENT.settings
        
    if xx is None:
        t_exp = settings['t_exp_'+inst]
        n_int = settings['t_exp_n_int_'+inst]
        model_flux = np.ones_like(config.BASEMENT.data[inst]['time']) #* np.nan               
    else:
        t_exp = None
        n_int = None
        model_flux = np.ones_like(xx) #* np.nan     
    
    
    #-------------------------------------------------------------------------- 
    #::: go through the time series transit by transit to fit for TTVs
    #-------------------------------------------------------------------------- 
    for n_transit in range(len(config.BASEMENT.data[companion+'_tmid_observed_transits'])):
        
        if xx is None:
            ind   = config.BASEMENT.data[inst][companion+'_ind_time_transit_'+str(n_transit+1)]
            xx_piecewise = config.BASEMENT.data[inst][companion+'_time_transit_'+str(n_transit+1)]
        else:
            tmid = config.BASEMENT.data[companion+'_tmid_observed_transits'][n_transit]
            width = settings['fast_fit_width']
            ind = np.where( (xx>=(tmid-width/2.)) \
                          & (xx<=(tmid+width/2.)) )[0]
            xx_piecewise = xx[ind]
        
        if len(xx_piecewise)>0:
            
            #::: FutureMax warning: one does not simply replace this with flux_subfct_ellc, because it need the additive term after epoch
            # model_flux_piecewise = flux_subfct_ellc(params, inst, companion, xx=xx_piecewise, settings=settings, t_exp=t_exp, n_int=n_int)

            #::: planet and EB transit lightcurve model
            if (params[companion+'_rr'] is not None) and (params[companion+'_rr'] > 0):
                model_flux_piecewise = ellc.lc(
                                  t_obs =       xx_piecewise, 
                                  radius_1 =    params[companion+'_radius_1'], 
                                  radius_2 =    params[companion+'_radius_2'], 
                                  sbratio =     params[companion+'_sbratio_'+inst], 
                                  incl =        params[companion+'_incl'], 
                                  light_3 =     params['dil_'+inst] / (1.-params['dil_'+inst]),
                                  t_zero =      params[companion+'_epoch'] + params[companion+'_ttv_transit_'+str(n_transit+1)],
                                  period =      params[companion+'_period'],
                                  a =           params[companion+'_a'],
                                  q =           params[companion+'_q'],
                                  f_c =         params[companion+'_f_c'],
                                  f_s =         params[companion+'_f_s'],
                                  ldc_1 =       params['host_ldc_'+inst],
                                  ldc_2 =       params[companion+'_ldc_'+inst],
                                  gdc_1 =       params['host_gdc_'+inst],
                                  gdc_2 =       params[companion+'_gdc_'+inst],
                                  didt =        params['didt_'+inst], 
                                  domdt =       params['domdt_'+inst], 
                                  rotfac_1 =    params['host_rotfac_'+inst], 
                                  rotfac_2 =    params[companion+'_rotfac_'+inst], 
                                  hf_1 =        params['host_hf_'+inst], #1.5, 
                                  hf_2 =        params[companion+'_hf_'+inst], #1.5,
                                  bfac_1 =      params['host_bfac_'+inst],
                                  bfac_2 =      params[companion+'_bfac_'+inst], 
                                  heat_1 =      divide(params['host_heat_'+inst],2.),
                                  heat_2 =      divide(params[companion+'_heat_'+inst],2.),
                                  lambda_1 =    params['host_lambda_'+inst], 
                                  lambda_2 =    params[companion+'_lambda_'+inst], 
                                  vsini_1 =     params['host_vsini'],
                                  vsini_2 =     params[companion+'_vsini'], 
                                  t_exp =       t_exp,
                                  n_int =       n_int,
                                  grid_1 =      settings['host_grid_'+inst],
                                  grid_2 =      settings[companion+'_grid_'+inst],
                                  ld_1 =        settings['host_ld_law_'+inst],
                                  ld_2 =        settings[companion+'_ld_law_'+inst],
                                  shape_1 =     settings['host_shape_'+inst],
                                  shape_2 =     settings[companion+'_shape_'+inst],
                                  spots_1 =     params['host_spots_'+inst], 
                                  spots_2 =     params[companion+'_spots_'+inst], 
                                  exact_grav =  settings['exact_grav'],
                                  verbose =     False
                                  )
                
                #::: and here comes an ugly hack around ellc, for those who want to fit reflected light and thermal emission separately
                '''
                if (companion+'_thermal_emission_amplitude_'+inst in params) and (params[companion+'_thermal_emission_amplitude_'+inst]>0):
                    model_flux += calc_thermal_curve(params, inst, companion, xx, t_exp, n_int)
                '''
                
            else:
                model_flux_piecewise = np.ones_like(xx)
                    
            model_flux[ind] = model_flux_piecewise
    

    #-------------------------------------------------------------------------- 
    #::: return
    #-------------------------------------------------------------------------- 
    return model_flux     




'''
#==============================================================================
#::: flux sub-fct: ellc phase curve hack 
#==============================================================================
#::: and here comes an ugly hack around ellc, for those who want to fit reflected light (i.e. geometric albedo) and thermal emission separately
def flux_subfct_ellc_phase_curve_hack(params, inst, companion, xx, t_exp, n_int):

    #::: a shift in the phase curve
    if (companion+'_thermal_emission_timeshift_'+inst in params):
        xx_shifted = xx - params[companion+'_thermal_emission_timeshift_'+inst]
    
    
    #::: the thermal curve evaluated at the requested time values (arbitrary scaling)
    occultation = ellc.lc( 
                      t_obs =       xx, 
                      radius_1 =    params[companion+'_radius_1'], 
                      radius_2 =    params[companion+'_radius_2'], 
                      sbratio =     1e12, #a hack to get the flux drop to 0 during occulation
                      incl =        params[companion+'_incl'], 
                      light_3 =     params['dil_'+inst] / (1.-params['dil_'+inst]),
                      t_zero =      params[companion+'_epoch'],
                      period =      params[companion+'_period'],
                      a =           params[companion+'_a'],
                      q =           params[companion+'_q'],
                      f_c =         params[companion+'_f_c'],
                      f_s =         params[companion+'_f_s'],
                      ldc_1 =       params['host_ldc_'+inst],
                      ldc_2 =       params[companion+'_ldc_'+inst],
                      gdc_1 =       0,
                      gdc_2 =       0,
                      didt =        params['didt_'+inst], 
                      domdt =       params['domdt_'+inst], 
                      rotfac_1 =    params['host_rotfac_'+inst], 
                      rotfac_2 =    params[companion+'_rotfac_'+inst], 
                      hf_1 =        params['host_hf_'+inst], #1.5, 
                      hf_2 =        params[companion+'_hf_'+inst], #1.5,
                      bfac_1 =      0,
                      bfac_2 =      0, 
                      heat_1 =      0,
                      heat_2 =      0,
                      lambda_1 =    params['host_lambda_'+inst], 
                      lambda_2 =    params[companion+'_lambda_'+inst], 
                      vsini_1 =     params['host_vsini'],
                      vsini_2 =     params[companion+'_vsini'], 
                      t_exp =       t_exp,
                      n_int =       n_int,
                      grid_1 =      config.BASEMENT.settings['host_grid_'+inst],
                      grid_2 =      config.BASEMENT.settings[companion+'_grid_'+inst],
                      ld_1 =        config.BASEMENT.settings['host_ld_law_'+inst],
                      ld_2 =        config.BASEMENT.settings[companion+'_ld_law_'+inst],
                      shape_1 =     'sphere',
                      shape_2 =     'sphere',
                      spots_1 =     None, 
                      spots_2 =     None, 
                      exact_grav =  config.BASEMENT.settings['exact_grav'],
                      verbose =     False
                      )
    
    #::: the thermal curve evaluated at the requested time values (arbitrary scaling)
    thermal_curve = ellc.lc( 
                      t_obs =       xx_shifted, 
                      radius_1 =    params[companion+'_radius_1'], 
                      radius_2 =    params[companion+'_radius_2'], 
                      sbratio =     0, 
                      incl =        params[companion+'_incl'], 
                      light_3 =     params['dil_'+inst] / (1.-params['dil_'+inst]),
                      t_zero =      params[companion+'_epoch'],
                      period =      params[companion+'_period'],
                      a =           params[companion+'_a'],
                      q =           params[companion+'_q'],
                      f_c =         params[companion+'_f_c'],
                      f_s =         params[companion+'_f_s'],
                      ldc_1 =       params['host_ldc_'+inst],
                      ldc_2 =       params[companion+'_ldc_'+inst],
                      gdc_1 =       0,
                      gdc_2 =       0,
                      didt =        params['didt_'+inst], 
                      domdt =       params['domdt_'+inst], 
                      rotfac_1 =    params['host_rotfac_'+inst], 
                      rotfac_2 =    params[companion+'_rotfac_'+inst], 
                      hf_1 =        params['host_hf_'+inst], #1.5, 
                      hf_2 =        params[companion+'_hf_'+inst], #1.5,
                      bfac_1 =      0,
                      bfac_2 =      0, 
                      heat_1 =      0,
                      heat_2 =      0.1,
                      lambda_1 =    params['host_lambda_'+inst], 
                      lambda_2 =    params[companion+'_lambda_'+inst], 
                      vsini_1 =     params['host_vsini'],
                      vsini_2 =     params[companion+'_vsini'], 
                      t_exp =       t_exp,
                      n_int =       n_int,
                      grid_1 =      config.BASEMENT.settings['host_grid_'+inst],
                      grid_2 =      config.BASEMENT.settings[companion+'_grid_'+inst],
                      ld_1 =        config.BASEMENT.settings['host_ld_law_'+inst],
                      ld_2 =        config.BASEMENT.settings[companion+'_ld_law_'+inst],
                      shape_1 =     'sphere',
                      shape_2 =     'sphere',
                      spots_1 =     None, 
                      spots_2 =     None, 
                      exact_grav =  config.BASEMENT.ssettings['exact_grav'],
                      verbose =     False
                      )
    
    #::: a finely sampled thermal curve (arbitray scaling; fine sampling to get the maximum)
    thermal_curve_fine = ellc.lc(
                      t_obs =       np.linspace(params[companion+'_epoch'], params[companion+'_epoch']+params[companion+'_period'], 1000), 
                      radius_1 =    params[companion+'_radius_1'], 
                      radius_2 =    params[companion+'_radius_2'], 
                      sbratio =     0,
                      incl =        params[companion+'_incl'], 
                      light_3 =     params['dil_'+inst] / (1.-params['dil_'+inst]),
                      t_zero =      params[companion+'_epoch'],
                      period =      params[companion+'_period'],
                      a =           params[companion+'_a'],
                      q =           params[companion+'_q'],
                      f_c =         params[companion+'_f_c'],
                      f_s =         params[companion+'_f_s'],
                      ldc_1 =       params['host_ldc_'+inst],
                      ldc_2 =       params[companion+'_ldc_'+inst],
                      gdc_1 =       0,
                      gdc_2 =       0,
                      didt =        params['didt_'+inst], 
                      domdt =       params['domdt_'+inst], 
                      rotfac_1 =    params['host_rotfac_'+inst], 
                      rotfac_2 =    params[companion+'_rotfac_'+inst], 
                      hf_1 =        params['host_hf_'+inst], #1.5, 
                      hf_2 =        params[companion+'_hf_'+inst], #1.5,
                      bfac_1 =      0,
                      bfac_2 =      0, 
                      heat_1 =      0,
                      heat_2 =      0.1,
                      lambda_1 =    params['host_lambda_'+inst], 
                      lambda_2 =    params[companion+'_lambda_'+inst], 
                      vsini_1 =     params['host_vsini'],
                      vsini_2 =     params[companion+'_vsini'], 
                      t_exp =       t_exp,
                      n_int =       n_int,
                      grid_1 =      config.BASEMENT.settings['host_grid_'+inst],
                      grid_2 =      config.BASEMENT.settings[companion+'_grid_'+inst],
                      ld_1 =        config.BASEMENT.settings['host_ld_law_'+inst],
                      ld_2 =        config.BASEMENT.settings[companion+'_ld_law_'+inst],
                      shape_1 =     'sphere',
                      shape_2 =     'sphere',
                      spots_1 =     None, 
                      spots_2 =     None, 
                      exact_grav =  config.BASEMENT.settings['exact_grav'],
                      verbose =     False
                      )
    
#    import matplotlib.pyplot as plt
#    
#    plt.figure()
#    plt.plot(xx_shifted, occultation)
#    
#    plt.figure()
#    plt.plot(xx_shifted, thermal_curve)
    
    #::: now scale the thermal curve
    thermal_curve[ thermal_curve<1. ] = 1.
    thermal_curve -= 1.
    thermal_curve *= occultation #a hack, he occultation time series is constant 0 in the occultation and constant 1 everywhere else
    thermal_curve /= np.max(thermal_curve_fine-1) #scaled from 0 to 1
    thermal_curve *= params[companion+'_thermal_emission_amplitude_'+inst]
    
    #::: cosine approximation
#            phi = calc_phase(xx_shifted, params[companion+'_period'], params[companion+'_epoch'])
#            thermal_curve += params[companion+'_thermal_emission_'+inst] * (0.5-0.5*np.cos(phi*2*np.pi))
    
    return thermal_curve
'''

    

###############################################################################
#::: rv fct
###############################################################################
def rv_fct(params, inst, companion, xx=None, settings=None):
    '''
    ! params must be updated via update_params() before calling this function !
    '''
    if xx is None:
        xx    = config.BASEMENT.data[inst]['time']
        t_exp = config.BASEMENT.settings['t_exp_'+inst]
        n_int = config.BASEMENT.settings['t_exp_n_int_'+inst]
    else:
        t_exp = None
        n_int = None
    
    if (params[companion+'_K'] is not None) and (params[companion+'_K'] > 0):
        model_rv1, model_rv2 = ellc.rv(
                          t_obs =       xx, 
                          radius_1 =    params[companion+'_radius_1'], 
                          radius_2 =    params[companion+'_radius_2'], 
                          sbratio =     params[companion+'_sbratio_'+inst], 
                          incl =        params[companion+'_incl'], 
                          t_zero =      params[companion+'_epoch'],
                          period =      params[companion+'_period'],
                          a =           params[companion+'_a'],
                          q =           params[companion+'_q'],
                          f_c =         params[companion+'_f_c'],
                          f_s =         params[companion+'_f_s'],
                          ldc_1 =       params['host_ldc_'+inst],
                          ldc_2 =       params[companion+'_ldc_'+inst],
                          gdc_1 =       params['host_gdc_'+inst],
                          gdc_2 =       params[companion+'_gdc_'+inst],
                          didt =        params['didt_'+inst], 
                          domdt =       params['domdt_'+inst], 
                          rotfac_1 =    params['host_rotfac_'+inst], 
                          rotfac_2 =    params[companion+'_rotfac_'+inst], 
                          hf_1 =        params['host_hf_'+inst], #1.5, 
                          hf_2 =        params[companion+'_hf_'+inst], #1.5,
                          bfac_1 =      params['host_bfac_'+inst],
                          bfac_2 =      params[companion+'_bfac_'+inst], 
                          heat_1 =      divide(params['host_heat_'+inst],2.),
                          heat_2 =      divide(params[companion+'_heat_'+inst],2.),
                          lambda_1 =    params['host_lambda_'+inst],
                          lambda_2 =    params[companion+'_lambda_'+inst], 
                          vsini_1 =     params['host_vsini'],
                          vsini_2 =     params[companion+'_vsini'], 
                          t_exp =       t_exp,
                          n_int =       n_int,
                          grid_1 =      config.BASEMENT.settings['host_grid_'+inst],
                          grid_2 =      config.BASEMENT.settings[companion+'_grid_'+inst],
                          ld_1 =        config.BASEMENT.settings['host_ld_law_'+inst],
                          ld_2 =        config.BASEMENT.settings[companion+'_ld_law_'+inst],
                          shape_1 =     config.BASEMENT.settings['host_shape_'+inst],
                          shape_2 =     config.BASEMENT.settings[companion+'_shape_'+inst],
                          spots_1 =     params['host_spots_'+inst], 
                          spots_2 =     params[companion+'_spots_'+inst], 
                          flux_weighted = config.BASEMENT.settings[companion+'_flux_weighted_'+inst],
                          verbose =     False
                          )
        
    else:
        model_rv1 = np.zeros_like(xx)
        model_rv2 = np.zeros_like(xx)
    
    return model_rv1, model_rv2




###############################################################################
#::: calculate external priors (e.g. stellar density and eccentricity cutoff)
###############################################################################  
def calculate_external_priors(params):
    '''
    ! params must be updated via update_params() before calling this function !
    
    bounds has to be list of len(theta), containing tuples of form
    ('none'), ('uniform', lower bound, upper bound), or ('normal', mean, std)
    '''
    lnp = 0.        
    
    #::: stellar density prior
    for companion in config.BASEMENT.settings['companions_phot']:
        '''
        The stellar density computed from R_host and M_host
        can be directly compared with the stellar density computed from the orbital motions (see e.g. Winn 210)
        '''
        if (config.BASEMENT.settings['use_host_density_prior'] is True) and ('host_density' in config.BASEMENT.external_priors) and (params[companion+'_host_density'] is not None):
            b = config.BASEMENT.external_priors['host_density']
            if b[0] == 'uniform':
                if not (b[1] <= params[companion+'_host_density'] <= b[2]): return -np.inf
            elif b[0] == 'normal':
                lnp += np.log( 1./(np.sqrt(2*np.pi) * b[2]) * np.exp( - (params[companion+'_host_density'] - b[1])**2 / (2.*b[2]**2) ) )
            else:
                raise ValueError('Bounds have to be "uniform" or "normal". Input was "'+b[0]+'".')
    
    #::: constrain eccentricities to avoid ellc crashes
    for companion in config.BASEMENT.settings['companions_all']:
        '''
        The EXOFASTv2 paper (https://arxiv.org/abs/1907.09480; page 37) introduces these constrains from collisions:
        ecc < 1 - (a+R2)/R1
        and from tidal limits (optional):
        ecc < 1 - 3a/R1
        '''
        #::: avoid runaway eccentricities
        if not (params[companion+'_ecc'] < 1.):
            lnp = -np.inf
        
        #::: avoid collisions
        if (params[companion+'_rsuma'] is not None) \
            and not ((params[companion+'_ecc'] < (1. - params[companion+'_rsuma']))): 
            lnp = -np.inf
            
        # ::: avoid tidal circularizaion
        if (params[companion+'_radius_1'] is not None) \
            and (config.BASEMENT.settings['use_tidal_eccentricity_prior'] is True) \
            and not (params[companion+'_ecc'] < (1. - 3*params[companion+'_radius_1'])): 
            lnp = -np.inf
            
    #::: constrain dilution to avoid ellc crashes
    for inst in config.BASEMENT.settings['inst_all']:
        if (params['dil_'+inst] > 0.999):
            lnp = -np.inf
            
    return lnp




###############################################################################
#::: calculate lnlike
###############################################################################  

#==============================================================================
#::: calculate all instruments linked (for stellar variability)
#==============================================================================
def calculate_lnlike_total(params):
#    print('\ncalculating lnlike total')
    
    lnlike_total = 0
    
    #--------------------------------------------------------------------------  
    #::: first, check and add external priors
    #--------------------------------------------------------------------------  
    lnprior_external = calculate_external_priors(params)   
    lnlike_total += lnprior_external       
    
            
    #--------------------------------------------------------------------------  
    #::: directly catch any issues
    #--------------------------------------------------------------------------  
    if np.isnan(lnlike_total) or np.isinf(lnlike_total):
        return -np.inf
    
    
    #--------------------------------------------------------------------------  
    #::: for all instruments
    #--------------------------------------------------------------------------   
    for key, key2 in zip(['flux','rv'], ['inst_phot', 'inst_rv']):      
               
#        print('\n',key)
#        print('a',config.BASEMENT.settings['stellar_var_'+key])
#        print('b',config.BASEMENT.settings['stellar_var_'+key] in FCTs)
        
#        print('c',[config.BASEMENT.settings['baseline_'+key+'_'+inst] for inst in config.BASEMENT.settings[key2]])
#        print('d',all( [config.BASEMENT.settings['baseline_'+key+'_'+inst] in FCTs for inst in config.BASEMENT.settings[key2]] ))
        
        #--------------------------------------------------------------------------       
        #::: CASE 1)
        #::: flux/rv stellar variability in FCTs --> can be calculated per inst (only GP needs to know about all other instruments) 
        #::: all flux/rv baselines in FCTs
        #-------------------------------------------------------------------------- 
        if ( config.BASEMENT.settings['stellar_var_'+key] in FCTs ) and all( [config.BASEMENT.settings['baseline_'+key+'_'+inst] in FCTs for inst in config.BASEMENT.settings[key2]] ):
#            print('CASE 1')
            for inst in config.BASEMENT.settings[key2]:
                
                #::: calculate the model; if there are any NaN, return -np.inf
                model = calculate_model(params, inst, key)
                if any(np.isnan(model)) or any(np.isinf(model)): 
                    return -np.inf
                
                #::: calculate errors, baseline and stellar variability
                yerr_w = calculate_yerr_w(params, inst, key)
                baseline = calculate_baseline(params, inst, key, model=model, yerr_w=yerr_w)
                stellar_var = calculate_stellar_var(params, inst, key, model=model, baseline=baseline, yerr_w=yerr_w)
                
                #::: calculate residuals and inv_simga2
                residuals = config.BASEMENT.data[inst][key] - model - baseline - stellar_var
                if any(np.isnan(residuals)): 
                    return -np.inf
#                    print(model)
#                    print(inst, key)
#                    print(yerr_w)
#                    print(baseline)
#                    print(stellar_var)
#                    for key in params:
#                        print(key, params[key])
#                    raise ValueError('There are NaN in the residuals. Something horrible happened.')
                inv_sigma2_w = 1./yerr_w**2
                
                #::: calculate lnlike
                lnlike_total += -0.5*(np.sum((residuals)**2 * inv_sigma2_w - np.log(inv_sigma2_w/2./np.pi))) #use np.sum to catch any nan and then set lnlike to nan
            
                
                
        #--------------------------------------------------------------------------  
        #::: CASES 2a) and 2b)
        #::: stellar variability in FCTs --> can be calculated per inst (only GP needs to know about all other instruments) 
        #::: baseline in FCTs or in GPs
        #::: then DO NOT calculate the residuals via flux - gp.predict
        #::: but use gp.log_likelihood instead
        #--------------------------------------------------------------------------  
        elif ( config.BASEMENT.settings['stellar_var_'+key] in FCTs ) and any( [config.BASEMENT.settings['baseline_'+key+'_'+inst] in GPs for inst in config.BASEMENT.settings[key2]] ):
#            print('CASE 2')
            for inst in config.BASEMENT.settings[key2]:
                
                #::: calculate the model; if there are any NaN, return -np.inf
                model = calculate_model(params, inst, key)
                if any(np.isnan(model)) or any(np.isinf(model)): return -np.inf
                
                
                #::: if that baseline is in FCTs
                if ( config.BASEMENT.settings['baseline_'+key+'_'+inst] in FCTs ):
#                    print('CASE 2a')
                    
                    #::: calculate errors, baseline and stellar variability
                    yerr_w = calculate_yerr_w(params, inst, key)
                    baseline = calculate_baseline(params, inst, key, model=model, yerr_w=yerr_w)
                    stellar_var = calculate_stellar_var(params, inst, key, model=model, baseline=baseline, yerr_w=yerr_w)
                    
                    #::: calculate residuals and inv_simga2
                    residuals = config.BASEMENT.data[inst][key] - model - baseline - stellar_var
                    if any(np.isnan(residuals)): raise ValueError('There are NaN in the residuals. Something horrible happened.')
                    inv_sigma2_w = 1./yerr_w**2
                    
                    #::: calculate lnlike
                    lnlike_total += -0.5*(np.sum((residuals)**2 * inv_sigma2_w - np.log(inv_sigma2_w/2./np.pi))) #use np.sum to catch any nan and then set lnlike to nan


                #::: if that baseline is in GPs
                elif ( config.BASEMENT.settings['baseline_'+key+'_'+inst] in GPs ):
#                    print('CASE 2b')
                
                    #::: calculate the errors and stellar variability (assuming baseline=0.)
                    yerr_w = calculate_yerr_w(params, inst, key)
                    stellar_var = calculate_stellar_var(params, inst, key, model=model, baseline=0., yerr_w=yerr_w)
                
                    #::: calculate the baseline's gp.log_likelihood (instead of evaluating the gp)
                    x = config.BASEMENT.data[inst]['time'] #pointer!
                    y = config.BASEMENT.data[inst][key] - model - stellar_var
                    gp = baseline_get_gp(params, inst, key)
                    try:
                        gp.compute(x, yerr=yerr_w)
                        lnlike_total += gp.log_likelihood(y)
                    except:
                        return -np.inf
                    
                    
                else:
                    raise ValueError('Kaput.')
                    
                    
                
        #--------------------------------------------------------------------------       
        #::: CASE 3) 
        #::: stellar variability in GPs
        #::: baseline in FCTs
        #::: do stuff
        #-------------------------------------------------------------------------- 
        elif ( config.BASEMENT.settings['stellar_var_'+key] in GPs ):
#            print('CASE 3')
            y, yerr_w = [], []
            for inst in config.BASEMENT.settings[key2]:
                
                #::: calculate the model. if there are any NaN, return -np.inf
                model_i = calculate_model(params, inst, key)
                if any(np.isnan(model_i)) or any(np.isinf(model_i)): return -np.inf
                               
                #::: calculate the errors and baseline
                yerr_w_i = calculate_yerr_w(params, inst, key)
                baseline_i = calculate_baseline(params, inst, key, model=model_i, yerr_w=yerr_w_i)
                residuals_i = config.BASEMENT.data[inst][key] - model_i - baseline_i

                y += list(residuals_i)
                yerr_w += list(yerr_w_i)
                
            #::: sort in time
            ind_sort = config.BASEMENT.data[key2]['ind_sort']
            x = 1.*config.BASEMENT.data[key2]['time']
            y = np.array(y)[ind_sort]
            yerr = np.array(yerr_w)[ind_sort]  
            
            #::: calculate the stellar variability's gp.log_likelihood (instead of evaluating the gp)
            gp = stellar_var_get_gp(params, key)
            try:
                gp.compute(x, yerr=yerr)
                lnlike_total += gp.log_likelihood(y)
            except:
                return -np.inf
            
            
                
        #--------------------------------------------------------------------------       
        #::: CASE 4) 
        #::: stellar variability in GPs
        #::: baseline in GPs
        #::: raise an error
        #--------------------------------------------------------------------------  
        elif ( config.BASEMENT.settings['stellar_var_'+key] in GPs )\
           and any( [config.BASEMENT.settings['baseline_'+key+'_'+inst] in GPs for inst in config.BASEMENT.settings[key2]] ):
            raise KeyError('Currently you cannot use a GP for stellar variability and a GP for the baseline at the same time.')

                    
            
    #--------------------------------------------------------------------------  
    #::: again, catch any issues
    #--------------------------------------------------------------------------  
    if np.isnan(lnlike_total) or np.isinf(lnlike_total):
        return -np.inf
        
        
    return lnlike_total


    
#==============================================================================
#::: calculate all instruments separately 
#::: (only possible if no stellar variability GP is involved)
#==============================================================================
#def calculate_lnlike(params, inst, key):
#        
#    #::: calculate the model. if there are any NaN, return -np.inf
#    model = calculate_model(params, inst, key)
#    if any(np.isnan(model)) or any(np.isinf(model)):
#        return -np.inf
#            
#    #--------------------------------------------------------------------------       
#    #::: CASE 3) 
#    #::: if no stellar variability GP and
#    #::: no baseline GP,
#    #::: then calculate lnlike manually
#    #--------------------------------------------------------------------------  
#    elif ( config.BASEMENT.settings['stellar_var_'+key] not in GPs ) \
#        and (config.BASEMENT.settings['baseline_'+key+'_'+inst] not in GPs ):
#        
#        yerr_w = calculate_yerr_w(params, inst, key)
#        baseline = calculate_baseline(params, inst, key, model=model, yerr_w=yerr_w)
#        
#        residuals = config.BASEMENT.data[inst][key] - model - baseline
#        inv_sigma2_w = 1./yerr_w**2
#        
#        lnlike = -0.5*(np.sum((residuals)**2 * inv_sigma2_w - np.log(inv_sigma2_w/2./np.pi))) #use np.sum to catch any nan and then set lnlike to nan
#    
#        if any(np.isnan(residuals)):
#            raise ValueError('There are NaN in the residuals. Something horrible happened.')
#    
#    
#    #--------------------------------------------------------------------------  
#    #::: CASE 4)
#    #::: if no stellar variability GP and
#    #::: baseline GP, 
#    #::: then DO NOT calculate the residuals via flux - gp.predict
#    #::: but use gp.log_likelihood instead
#    #--------------------------------------------------------------------------  
#    else:
#        x = config.BASEMENT.data[inst]['time']
#        y = config.BASEMENT.data[inst][key] - model
#        yerr_w = calculate_yerr_w(params, inst, key)
#        gp = baseline_get_gp(params, inst, key)
#        try:
#            gp.compute(x, yerr=yerr_w)
#            lnlike = gp.log_likelihood(y)
#        except:
#            lnlike = -np.inf
#        
#        
#    return lnlike
    
    
    

###############################################################################
#::: calculate yerr
############################################################################### 
def calculate_yerr_w(params, inst, key):
    '''
    Returns:
    --------
    yerr_w : array of float
        the weighted yerr
    '''
    if inst in config.BASEMENT.settings['inst_phot']:
        yerr_w = config.BASEMENT.data[inst]['err_scales_'+key] * params['err_'+key+'_'+inst]
    elif inst in config.BASEMENT.settings['inst_rv']:
        yerr_w = np.sqrt( config.BASEMENT.data[inst]['white_noise_'+key]**2 + params['jitter_'+key+'_'+inst]**2 )
    return yerr_w


        

################################################################################
##::: calculate residuals
################################################################################  
def calculate_residuals(params, inst, key):
    '''
    Note:
    -----
    No 'xx' here, because residuals can only be calculated on given data
    (not on an arbitrary xx grid)
    '''       
    model = calculate_model(params, inst, key)
    baseline = calculate_baseline(params, inst, key, model=model)
    residuals = config.BASEMENT.data[inst][key] - model - baseline
    return residuals


    

###############################################################################
#::: calculate model
###############################################################################      
def calculate_model(params, inst, key, xx=None, settings=None):
            
    if settings is None:
        settings = config.BASEMENT.settings
    
    if key=='flux':
        depth = 0.
        for companion in settings['companions_phot']:
            depth += ( 1. - flux_fct(params, inst, companion, xx=xx, settings=settings) )
        model_flux = 1. - depth
        return model_flux
    
    elif key=='rv':
        model_rv = 0.
        for companion in settings['companions_rv']:
            model_rv += rv_fct(params, inst, companion, xx=xx, settings=settings)[0]
        return model_rv
    
    elif (key=='centdx') | (key=='centdy'):
        raise ValueError("Fitting for 'centdx' and 'centdy' not yet implemented.")
        #TODO
        
    else:
        raise ValueError("Variable 'key' has to be 'flux', 'rv', 'centdx', or 'centdy'.")




###############################################################################
#::: calculate baseline
############################################################################### 
        
#==============================================================================
#::: calculate baseline: main
#==============================================================================
def calculate_baseline(params, inst, key, model=None, yerr_w=None, xx=None):

    '''
    Inputs:
    -------
    params : dict
        ...
    inst : str
        ...
    key : str
        ...
    model = array of float (optional; default=None)
        ...
    xx : array of float (optional; default=None)
        if given, evaluate the baseline fit on the xx values 
        (e.g. a finer time grid for plotting)
        else, it's the same as data[inst]['time']
        
    Returns: 
    --------
    baseline : array of float
        the baseline evaluate on the grid x (or xx, if xx!=None)
    '''
    
    if model is None: 
        model = calculate_model(params, inst, key, xx=None) #the model has to be evaluated on the time grid
    if yerr_w is None: 
        yerr_w = calculate_yerr_w(params, inst, key)
    x = config.BASEMENT.data[inst]['time']
    y = config.BASEMENT.data[inst][key] - model
    if xx is None:  
        xx = 1.*x
    '''
    x : array of float
        time stamps of the data
    y : array of float
        y = data_y - model_y
        i.e., the values that you want to constrain the baseline on
    yerr_w : array of float
        the weighted yerr
    yerr_weights : array of float
        normalized error weights on y
    '''
    
    baseline_method = config.BASEMENT.settings['baseline_'+key+'_'+inst]
    
    return baseline_switch[baseline_method](x, y, yerr_w, xx, params, inst, key)



#==============================================================================
#::: calculate baseline: hybrid_offset (like Gillon+2012, but only remove mean offset)
#==============================================================================
def baseline_hybrid_offset(*args):
    x, y, yerr_w, xx, params, inst, key = args
    yerr_weights = yerr_w/np.nanmean(yerr_w)
    weights = 1./yerr_weights
    ind = np.isfinite(y) #np.average can't handle NaN
    return np.average(y[ind], weights=weights[ind]) * np.ones_like(xx)
 

    
#==============================================================================
#::: calculate baseline: hybrid_poly (like Gillon+2012)
#==============================================================================   
def baseline_hybrid_poly(*args):
    x, y, yerr_w, xx, params, inst, key = args
    polyorder = int(config.BASEMENT.settings['baseline_'+key+'_'+inst][-1])
    xx = (xx - x[0])/x[-1] #polyfit needs the xx-axis scaled to [0,1], otherwise it goes nuts
    x = (x - x[0])/x[-1] #polyfit needs the x-axis scaled to [0,1], otherwise it goes nuts
    if polyorder>=0:
        yerr_weights = yerr_w/np.nanmean(yerr_w)
        weights = 1./yerr_weights
        ind = np.isfinite(y) #polyfit can't handle NaN
        params_poly = poly.polyfit(x[ind],y[ind],polyorder,w=weights[ind]) #WARNING: returns params in reverse order than np.polyfit!!!
        baseline = poly.polyval(xx, params_poly) #evaluate on xx (!)
    else:
        raise ValueError("'polyorder' has to be > 0.")
    return baseline    



#==============================================================================
#::: calculate baseline: hybrid_spline (like Gillon+2012, but with a cubic spline)
#==============================================================================
def baseline_hybrid_spline(*args):
    x, y, yerr_w, xx, params, inst, key = args
    yerr_weights = yerr_w/np.nanmean(yerr_w)
    weights = 1./yerr_weights
    ind = np.isfinite(y) #mask NaN
    spl = UnivariateSpline(x[ind],y[ind],w=weights[ind],s=np.sum(weights[ind]))
    baseline = spl(xx)
    
#    if any(np.isnan(baseline)):
#        import matplotlib.pyplot as plt
#        print(x[ind])
#        print(y[ind])
#        print(weights[ind])
#        plt.figure()
#        plt.plot(x,y,'k.', color='grey')
#        plt.plot(xx,baseline,'r-', lw=2)
#        plt.show()
#        input('press enter to continue')
    
    return baseline   

     
    
#==============================================================================
#::: calculate baseline: hybrid_GP (like Gillon+2012, but with a GP)
#==============================================================================           
def baseline_hybrid_GP(*args):
    x, y, yerr_w, xx, params, inst, key = args
    
    kernel = terms.Matern32Term(log_sigma=1., log_rho=1.)
    gp = celerite.GP(kernel, mean=np.nanmean(y)) 
    gp.compute(x, yerr=yerr_w) #constrain on x/y/yerr
     
    def neg_log_like(gp_params, y, gp):
        gp.set_parameter_vector(gp_params)
        return -gp.log_likelihood(y)
    
    def grad_neg_log_like(gp_params, y, gp):
        gp.set_parameter_vector(gp_params)
        return -gp.grad_log_likelihood(y)[1]
    
    initial_params = gp.get_parameter_vector()
    bounds = gp.get_parameter_bounds()
    soln = minimize(neg_log_like, initial_params, jac=grad_neg_log_like,
                    method="L-BFGS-B", bounds=bounds, args=(y, gp))
    gp.set_parameter_vector(soln.x)
    
    baseline = gp_predict_in_chunks(gp, y, xx)[0]
#    baseline = gp.predict(y, xx)[0] #constrain on x/y/yerr, evaluate on xx (!)
    return baseline 



#==============================================================================
#::: calculate baseline: sample_offset
########################################################################### 
def baseline_sample_offset(*args):
    x, y, yerr_w, xx, params, inst, key = args
    return params['baseline_offset_'+key+'_'+inst] * np.ones_like(xx)
        


#==============================================================================
#::: calculate baseline: sample_linear
#============================================================================== 
def baseline_sample_linear(*args):
    x, y, yerr_w, xx, params, inst, key = args
    xx_norm = (xx-x[0]) / (x[-1]-x[0])
    return params['baseline_slope_'+key+'_'+inst] * xx_norm + params['baseline_offset_'+key+'_'+inst]
        
    
    
#==============================================================================
#::: calculate baseline: sample_GP
#============================================================================== 
def baseline_sample_GP(*args):
    x, y, yerr_w, xx, params, inst, key = args
    gp = baseline_get_gp(params, inst, key)
    gp.compute(x, yerr=yerr_w)
    baseline = gp_predict_in_chunks(gp, y, xx)[0]
#    baseline = gp.predict(y, xx)[0]
    
#    baseline2 = gp.predict(y, x)[0]
#    plt.figure()
#    plt.plot(x,y,'k.', color='grey')
#    plt.plot(xx,baseline,'r-', lw=2)
#    plt.plot(x,baseline2,'ro', lw=2)
#    plt.title(inst+' '+key+' '+str(gp.get_parameter_vector()))
#    plt.show()
#    raw_input('press enter to continue')
    
    return baseline



#==============================================================================
#::: calculate baseline: get GP kernel
#============================================================================== 
def baseline_get_gp(params, inst, key):
    
    #::: kernel
    if config.BASEMENT.settings['baseline_'+key+'_'+inst] == 'sample_GP_real':
        kernel = terms.RealTerm(log_a=params['baseline_gp_real_lna_'+key+'_'+inst], 
                                log_c=params['baseline_gp_real_lnc_'+key+'_'+inst])
        
    elif config.BASEMENT.settings['baseline_'+key+'_'+inst] == 'sample_GP_complex':
        kernel = terms.ComplexTerm(log_a=params['baseline_gp_complex_lna_'+key+'_'+inst], 
                                log_b=params['baseline_gp_complex_lnb_'+key+'_'+inst], 
                                log_c=params['baseline_gp_complex_lnc_'+key+'_'+inst], 
                                log_d=params['baseline_gp_complex_lnd_'+key+'_'+inst])
                  
    elif config.BASEMENT.settings['baseline_'+key+'_'+inst] == 'sample_GP_Matern32':
        kernel = terms.Matern32Term(log_sigma=params['baseline_gp_matern32_lnsigma_'+key+'_'+inst], 
                                    log_rho=params['baseline_gp_matern32_lnrho_'+key+'_'+inst])
        
    elif config.BASEMENT.settings['baseline_'+key+'_'+inst] == 'sample_GP_SHO':
        kernel = terms.SHOTerm(log_S0=params['baseline_gp_sho_lnS0_'+key+'_'+inst], 
                               log_Q=params['baseline_gp_sho_lnQ_'+key+'_'+inst],
                               log_omega0=params['baseline_gp_sho_lnomega0_'+key+'_'+inst])                               
        
    else: 
        KeyError('GP settings and params do not match.')
    
    #::: GP and mean (simple offset)  
    if 'baseline_gp_offset_'+key+'_'+inst in params:
        gp = celerite.GP(kernel, mean=params['baseline_gp_offset_'+key+'_'+inst], fit_mean=True)
    else:
        gp = celerite.GP(kernel, mean=0.) #mean=0. because it is the mean of the residuals, as the GP is applied to the residuals
        
    return gp



#==============================================================================
#::: calculate baseline: none
#============================================================================== 
def baseline_none(*args):
    x, y, yerr_w, xx, params, inst, key = args
    return np.zeros_like(xx)



#==============================================================================
#::: calculate baseline: raise error
#==============================================================================
def baseline_raise_error(*args):
    x, y, yerr_w, xx, params, inst, key = args
    raise ValueError('Setting '+'baseline_'+key+'_'+inst+' has to be sample_offset / sample_linear / sample_GP / hybrid_offset / hybrid_poly_1 / hybrid_poly_2 / hybrid_poly_3 / hybrid_pol_4 / hybrid_spline / hybrid_GP, '+\
                     "\nbut is:"+config.BASEMENT.settings['baseline_'+key+'_'+inst])



#==============================================================================
#::: calculate baseline: baseline_switch
#==============================================================================    
baseline_switch = \
    {
    'hybrid_offset' : baseline_hybrid_offset,
    'hybrid_poly_0' : baseline_hybrid_poly,
    'hybrid_poly_1' : baseline_hybrid_poly,
    'hybrid_poly_2' : baseline_hybrid_poly,
    'hybrid_poly_3' : baseline_hybrid_poly,
    'hybrid_poly_4' : baseline_hybrid_poly,
    'hybrid_poly_5' : baseline_hybrid_poly,
    'hybrid_poly_6' : baseline_hybrid_poly,
    'hybrid_spline' : baseline_hybrid_spline,
    'hybrid_GP'     : baseline_hybrid_GP,
    'sample_offset' : baseline_sample_offset,
    'sample_linear' : baseline_sample_linear,
    'sample_GP_real'         : baseline_sample_GP, #only for plotting
    'sample_GP_complex'      : baseline_sample_GP, #only for plotting   
    'sample_GP_Matern32'     : baseline_sample_GP, #only for plotting
    'sample_GP_SHO'          : baseline_sample_GP, #only for plotting         
    'none'          : baseline_none #only for plotting    
    }
    

    
    
###########################################################################
#::: GP predict in chunks (to avoid memory crashes)
########################################################################### 
def gp_predict_in_chunks(gp, y, x, chunk_size=5000):
    mu = []
    var = []
    for i in range( int(1.*len(x)/chunk_size)+1 ):
        m, v = gp.predict(y, x[i*chunk_size:(i+1)*chunk_size], return_var=True)
        mu += list(m)
        var += list(v)
    return np.array(mu), np.array(var)




###############################################################################
#::: Stellar Variability
############################################################################### 
            
#==============================================================================
#::: Stellar Variability: main
#==============================================================================
def calculate_stellar_var(params, inst, key, model=None, baseline=None, yerr_w=None, xx=None):

    #--------------------------------------------------------------------------
    #::: over all instruments (needed for GP)
    #--------------------------------------------------------------------------
    stellar_var_method = config.BASEMENT.settings['stellar_var_'+key]   
    
    if stellar_var_method not in ['none']:
        if key=='flux': key2 = 'inst_phot'
        elif key=='rv': key2 = 'inst_rv'
        else: KeyError('Kaput.')
        
        if inst=='all': insts = config.BASEMENT.settings[key2]
        else: insts = [inst]
        
        y_list,yerr_w_list = [],[]
        for inst in insts:
            if model is None: 
                model_i = calculate_model(params, inst, key)
            else:
                model_i = model
            if baseline is None: 
                baseline_i = calculate_baseline(params, inst, key, model=model)
            else:
                baseline_i = baseline
            residuals = config.BASEMENT.data[inst][key] - model_i - baseline_i
            y_list += list(residuals)
            
            if yerr_w is None: 
                yerr_w_list += list(calculate_yerr_w(params, inst, key))
            else: 
                yerr_w_list += list(yerr_w)
              
        if inst=='all': ind_sort = config.BASEMENT.data[key2]['ind_sort']
        else: ind_sort = slice(None)
        
        x = 1.*config.BASEMENT.data[key2]['time']
        y = np.array(y_list)[ind_sort]
        yerr_w = np.array(yerr_w_list)[ind_sort]  
        if xx is None: xx = 1.*x
    
        return stellar_var_switch[stellar_var_method](x, y, yerr_w, xx, params, key)
    
    else:
        return 0.
    


#==============================================================================
#::: Stellar Variability: get GP kernel
#============================================================================== 
def stellar_var_get_gp(params, key):
    
    #::: kernel
    if config.BASEMENT.settings['stellar_var_'+key] == 'sample_GP_real':
        kernel = terms.RealTerm(log_a=params['stellar_var_gp_real_lna_'+key], 
                                log_c=params['stellar_var_gp_real_lnc_'+key])
        
    elif config.BASEMENT.settings['stellar_var_'+key] == 'sample_GP_complex':
        kernel = terms.ComplexTerm(log_a=params['stellar_var_gp_complex_lna_'+key], 
                                log_b=params['stellar_var_gp_complex_lnb_'+key], 
                                log_c=params['stellar_var_gp_complex_lnc_'+key], 
                                log_d=params['stellar_var_gp_complex_lnd_'+key])
                  
    elif config.BASEMENT.settings['stellar_var_'+key] == 'sample_GP_Matern32':
        kernel = terms.Matern32Term(log_sigma=params['stellar_var_gp_matern32_lnsigma_'+key], 
                                    log_rho=params['stellar_var_gp_matern32_lnrho_'+key])
        
    elif config.BASEMENT.settings['stellar_var_'+key] == 'sample_GP_SHO':
        kernel = terms.SHOTerm(log_S0=params['stellar_var_gp_sho_lnS0_'+key], 
                               log_Q=params['stellar_var_gp_sho_lnQ_'+key],
                               log_omega0=params['stellar_var_gp_sho_lnomega0_'+key])                               
        
    else: 
        KeyError('GP settings and params do not match.')
    
    #::: GP and mean (simple offset)  
    if 'stellar_var_gp_offset_'+key in params:
        gp = celerite.GP(kernel, mean=params['stellar_var_gp_offset_'+key], fit_mean=True)
    else:
        gp = celerite.GP(kernel, mean=0.) #mean=0. because it is the mean of the residuals, as the GP is applied to the residuals
        
        
    return gp



#==============================================================================
#::: Stellar Variability: sample_GP
#============================================================================== 
def stellar_var_sample_GP(*args):
    x, y, yerr_w, xx, params, key = args
    gp = stellar_var_get_gp(params, key)
    gp.compute(x, yerr=yerr_w)
    stellar_var = gp_predict_in_chunks(gp, y, xx)[0]
    return stellar_var



#==============================================================================
#::: Stellar Variability: sample_linear
#============================================================================== 
def stellar_var_sample_linear(*args):
    x, y, yerr_w, xx, params, key = args
    if key=='flux': key2 = 'inst_phot'
    elif key=='rv': key2 = 'inst_rv'
    x_all = 1.*config.BASEMENT.data[key2]['time']
    xx_norm = (xx-x_all[0]) / (x_all[-1]-x_all[0])
#    xx_norm = (xx - config.BASEMENT.settings['mid_epoch'])
#    xx_norm = (xx - 2457000.) / 1000.
    return params['stellar_var_slope_'+key] * xx_norm + params['stellar_var_offset_'+key]
        
    

#==============================================================================
#::: Stellar Variability: none
#============================================================================== 
def stellar_var_none(*args):
    x, y, yerr_w, xx, params, key = args
    return np.zeros_like(xx)



#==============================================================================
#::: Stellar Variability: stellar_var_switch
#==============================================================================    
stellar_var_switch = \
    {
    'sample_linear'          : stellar_var_sample_linear,
    'sample_GP_real'         : stellar_var_sample_GP, #only for plotting
    'sample_GP_complex'      : stellar_var_sample_GP, #only for plotting   
    'sample_GP_Matern32'     : stellar_var_sample_GP, #only for plotting
    'sample_GP_SHO'          : stellar_var_sample_GP, #only for plotting           
    'none'                   : stellar_var_none #only for plotting    
    }
    


    
################################################################################
##::: def calculate inv_sigma2
################################################################################  
#def calculate_inv_sigma2_w(params, inst, key, residuals=None):
#    '''
#    _w means "weighted", a.k.a. multiplied by data[inst]['err_scales_'+key]**(-2)
#    '''
#    
#    #::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::    
#    #::: traditional (sampling in MCMC)
#    #::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::  
#    if config.BASEMENT.settings['error_'+key+'_'+inst].lower() == 'sample':
#        yerr_w = calculate_yerr_w(params, inst, key)
#        inv_sigma2_w = 1./yerr_w**2
#        return inv_sigma2_w
#    
#     
#    #:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::: 
#    #::: 'hybrid_inv_sigma2'
#    #:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::: 
#    elif config.BASEMENT.settings['error_'+key+'_'+inst].lower() == 'hybrid': 
#        raise ValueError('Currently no longer implemented.')
##        if residuals is None:
##            residuals = calculate_residuals(params, inst, key)
##        
##        #::: neg log like
##        def neg_log_like(inv_sigma2, inst, key, residuals):
###            inv_sigma2_w = config.BASEMENT.data[inst]['err_scales_'+key]**(-2) * inv_sigma2
##            inv_sigma2_w = calculate_inv_sigma2_w_1(inv_sigma2, inst, key)                
##            return + 0.5*(np.nansum((residuals)**2 * inv_sigma2_w - np.log(inv_sigma2_w)))
##            
##        
##        #::: grad neg log like
##        def grad_neg_log_like(inv_sigma2, inst, key, residuals):
###            inv_sigma2_w = config.BASEMENT.data[inst]['err_scales_'+key]**(-2) * inv_sigma2
##            inv_sigma2_w = calculate_inv_sigma2_w_1(inv_sigma2, inst, key)                
##            return np.array( + 0.5*(np.nansum((residuals)**2 - 1./inv_sigma2_w)) )
##        
##        
###        guess = params['inv_sigma2_'+key+'_'+inst]
##        guess = 1./np.std(residuals)**2 #Warning: this relies on a good initial guess for the model, otherwise std(residuals) will be nuts
##
##        #::: MLE (gradient based)
##        soln_MLE = minimize(neg_log_like, guess,
##                        method = 'L-BFGS-B', jac=grad_neg_log_like,
##                        bounds=[(1e-16,1e+16)], args=(inst, key, residuals))
##        
##        #::: Diff. Evol.
###        bounds = [(0.001*guess,1000.*guess)]
###        soln_DE = differential_evolution(neg_log_like, bounds, args=(inst, key, residuals))
##
##        inv_sigma2 = soln_MLE.x[0]      
##        inv_sigma2_w = calculate_inv_sigma2_w_1(inv_sigma2, inst, key)                
##
###        print inst, key
###        print '\tguess:', int(guess), 'lnlike:', neg_log_like(guess, inst, key, residuals)
###        print '\tMLE:', int(soln_MLE.x[0]), 'lnlike:', neg_log_like(soln_MLE.x[0], inst, key, residuals) 
###        print '\tDE:', int(soln_DE.x[0]), 'lnlike:', neg_log_like(soln_DE.x[0], inst, key, residuals)
##        
##        return inv_sigma2_w
#    
#
#    else:
#        raise ValueError('Setting '+'error_'+key+'_'+inst+' has to be sample / hybrid, '+\
#                         "\nbut is:"+params['error_'+key+'_'+inst])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Jun 29 20:44:29 2020

@author:
Maximilian N. Günther
MIT Kavli Institute for Astrophysics and Space Research, 
Massachusetts Institute of Technology,
77 Massachusetts Avenue,
Cambridge, MA 02109, 
USA
Email: maxgue@mit.edu
Web: www.mnguenther.com
"""

from __future__ import print_function, division, absolute_import

#::: modules
import os, sys
import numpy as np
import matplotlib.pyplot as plt
import pandas as pd
from tqdm import tqdm
from glob import glob
from pprint import pprint
from wotan import flatten
from astropy.stats import sigma_clip

#::: my modules
import allesfitter

#::: plotting settings
import seaborn as sns
sns.set(context='paper', style='ticks', palette='deep', font='sans-serif', font_scale=1.5, color_codes=True)
sns.set_style({"xtick.direction": "in","ytick.direction": "in"})
sns.set_context(rc={'lines.markeredgewidth': 1})




def fast_slide_clip(time, flux, window_length=1, low=5, high=5):
    """
    A much faster alternative to Wotan's build-in slide clip

    Returns
    -------
    Clipped flux (outliers replaced with NaN)
    """
    flux2 = 1.*flux
    flux_flat = flatten(time, flux, method='biweight', window_length=window_length)
    flux_flat = sigma_clip(np.ma.masked_invalid(flux_flat), sigma_lower=low, sigma_upper=high) #astropy wants masked arrays
    flux2[flux_flat.mask] = np.nan #use NaN instead of masked arrays, because masked arrays drive me crazy
    return flux2


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Mar  5 17:22:59 2020

@author: 
Dr. Maximilian N. Günther
MIT Kavli Institute for Astrophysics and Space Research, 
Massachusetts Institute of Technology,
77 Massachusetts Avenue,
Cambridge, MA 02109, 
USA
Email: maxgue@mit.edu
GitHub: https://github.com/MNGuenther
Web: www.mnguenther.com
"""

from __future__ import print_function, division, absolute_import

#::: modules
import numpy as np
import matplotlib.pyplot as plt
import os, sys

#::: my modules
from .translator import translate

#::: plotting settings
import seaborn as sns
sns.set(context='paper', style='ticks', palette='deep', font='sans-serif', font_scale=1.5, color_codes=True)
sns.set_style({"xtick.direction": "in","ytick.direction": "in"})
sns.set_context(rc={'lines.markeredgewidth': 1})




'''
A collection of plug-and-play plotting functions, to add to axes in a gridspec or subplot layout
'''
    
    

    
###############################################################################
#::: full lightcurve
###############################################################################
def plot_lc_full(ax, time, flux, flux_err, time_grid, model_flux_grid):
    ax.errorbar(time, flux, yerr=flux_err, fmt='b.', rasterized=True, zorder=11)
    ax.plot(time_grid, model_flux_grid, 'r-', lw=2, zorder=12)
    ax.set(xlabel='Time'+r' $\mathrm{(BJD_{TDB})}$', ylabel='Flux')
    return ax


###############################################################################
#::: phase-folded lightcurve
###############################################################################
def plot_lc_phase(ax, phi, flux, phase, phaseflux, phaseflux_err, phase_grid, model_phaseflux_grid):
    ax.plot(phi, flux, 'k.', color='lightgrey', rasterized=True, zorder=10)
    ax.errorbar(phase, phaseflux, yerr=phaseflux_err, fmt='b.', zorder=11)
    ax.plot(phase_grid, model_phaseflux_grid, 'r-', lw=2, zorder=12)
    ax.set(xlabel='Phase', ylabel='Flux')
    return ax


###############################################################################
#::: phase-folded and zoomed lightcurve; same as for plot_phase(), but zoomed in x and in minutes
###############################################################################
def plot_lc_phasezoom(ax, phi, flux, phase, phaseflux, phaseflux_err, phase_grid, model_phaseflux_grid, zoomfactor):
    '''
    zoomfactor : period * 24 * 60
    '''
    ax.plot(phi*zoomfactor, flux, 'k.', color='lightgrey', rasterized=True, zorder=10)
    ax.errorbar(phase*zoomfactor, phaseflux, yerr=phaseflux_err, fmt='b.', zorder=11)
    ax.plot(phase_grid*zoomfactor, model_phaseflux_grid, 'r-', lw=2, zorder=12)
    ax.set(xlim=[-240,240])
    ax.set(xlabel='Time (min.)', ylabel='Flux')
    return ax
    
    
###############################################################################
#::: full RV series
###############################################################################
def plot_rv_full(ax, time, rv, rv_err, time_grid, model_flux_grid):
    ax.errorbar(time, rv, yerr=rv_err, fmt='bo', zorder=11)
    ax.plot(time_grid, model_flux_grid, 'r-', lw=2, zorder=12)
    ax.set(xlabel='Time'+r' $\mathrm{(BJD_{TDB})}$', ylabel='RV (km/s)')
    return ax


###############################################################################
#::: phase-folded RV series
###############################################################################
def plot_rv_phase(ax, phi, rv, rv_err, phase_grid, model_phaserv_grid):
    ax.errorbar(phi, rv, yerr=rv_err, fmt='bo', zorder=11)
    ax.plot(phase_grid, model_phaserv_grid, 'r-', lw=2, zorder=12)
    ax.set(xlabel='Phase', ylabel='RV (km/s)')
    return ax


###############################################################################
#::: info text 
###############################################################################
def plot_info(ax, text=0, params=None, settings=None, **kwargs):
    params = translate(params=params, settings=settings, quiet=True, **kwargs)
    
    ax.set_axis_off()
    
    if text==0:
        ax.text(0,0.95,'R_comp = '+np.format_float_positional(params['R_companion_earth'],2)+r' $R_\odot$ = ' + np.format_float_positional(params['R_companion_jup'],2)+r' $R_J$ = ' + np.format_float_positional(params['R_companion_sun'],2)+r' $R_\odot$', transform=ax.transAxes) #bodies
        ax.text(0,0.85,'M_comp = '+np.format_float_positional(params['M_companion_earth'],2)+r' $M_\odot$ = ' + np.format_float_positional(params['M_companion_jup'],2)+r' $M_J$ = ' + np.format_float_positional(params['M_companion_sun'],2)+r' $M_\odot$', transform=ax.transAxes)
        ax.text(0,0.75,'R_host = '+str(params['R_host'])+r' $R_\odot$', transform=ax.transAxes)
        ax.text(0,0.65,'M_host = '+str(params['M_host'])+r' $M_\odot$', transform=ax.transAxes)
        ax.text(0,0.55,'sbratio = '+str(params['sbratio']), transform=ax.transAxes)
        ax.text(0,0.45,'epoch' + ' = '+str(params['epoch'])+r' $\mathrm{BJD_{TDB}}$', transform=ax.transAxes) #orbits
        ax.text(0,0.35,'period = '+str(params['period'])+' days', transform=ax.transAxes)
        ax.text(0,0.25,'incl = '+str(params['incl'])+' deg', transform=ax.transAxes)
        ax.text(0,0.15,'ecc = '+str(params['ecc']), transform=ax.transAxes)
        ax.text(0,0.05,'omega = '+str(params['omega'])+' deg', transform=ax.transAxes)
    
    if text==1:
        ax.text(0,0.95,'dil = '+str(params['dil']), transform=ax.transAxes)
        ax.text(0,0.85,'R_comp/R_host = '+np.format_float_positional(params['rr'],5,False), transform=ax.transAxes)
        ax.text(0,0.75,'(R_comp+R_host)/a = '+np.format_float_positional(params['rsuma'],5,False), transform=ax.transAxes)
        ax.text(0,0.65,'R_comp/a = '+np.format_float_positional(params['R_companion_over_a'],5,False), transform=ax.transAxes)
        ax.text(0,0.55,'R_host/a = '+np.format_float_positional(params['R_host_over_a'],5,False), transform=ax.transAxes)
        ax.text(0,0.45,'cosi = '+np.format_float_positional(params['cosi'],5,False), transform=ax.transAxes)
        ax.text(0,0.35,r'$\sqrt{e} \cos{\omega}$ = '+np.format_float_positional(params['f_c'],5,False), transform=ax.transAxes)
        ax.text(0,0.25,r'$\sqrt{e} \sin{\omega}$ = '+np.format_float_positional(params['f_s'],5,False), transform=ax.transAxes)
        ax.text(0,0.15,'LD = '+str(params['ldc']), transform=ax.transAxes)
        try: ax.text(0,0.05,'LD transf = ['+", ".join([np.format_float_positional(item,5,False) for item in params['ldc_transformed']]) + ']', transform=ax.transAxes)
        except: pass
    return ax


#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""
Created on Tue Oct  2 10:54:58 2018

@author:
Dr. Maximilian N. Guenther
MIT Kavli Institute for Astrophysics and Space Research, 
Massachusetts Institute of Technology,
77 Massachusetts Avenue,
Cambridge, MA 02109, 
USA
Email: maxgue@mit.edu
Web: www.mnguenther.com
"""

#::: plotting settings
import seaborn as sns
sns.set(context='paper', style='ticks', palette='deep', font='sans-serif', font_scale=1.5, color_codes=True)
sns.set_style({"xtick.direction": "in","ytick.direction": "in"})
sns.set_context(rc={'lines.markeredgewidth': 1})

#::: modules
import numpy as np
import matplotlib.pyplot as plt
from scipy.optimize import minimize
from scipy.stats import skewnorm




np.random.seed(42)




def simulate_PDF(median, lower_err, upper_err, size=1, plot=True):
    '''
    Simulates a draw of posterior samples from a value and asymmetric errorbars
    by assuming the underlying distribution is a skewed normal distribution.
    
    Developed to estimate PDFs from literature exoplanet parameters that did not report their MCMC chains.
    
    Inputs:
    -------
    median : float
        the median value that was reported
    lower_err : float
        the lower errorbar that was reported
    upper_err : float
        the upper errorbar that was reported
    size : int
        the number of samples to be drawn
        
    Returns:
    --------
    samples : array of float
        the samples drawn from the simulated skrewed normal distribution
    '''
    
    sigma, omega, alpha = calculate_skewed_normal_params(median, lower_err, upper_err)
    samples = skewnorm.rvs(alpha, loc=sigma, scale=omega, size=size)
    
    if plot==False:
        return samples

    else:
        lower_err = np.abs(lower_err)
        upper_err = np.abs(upper_err)
        x = np.arange(median-4*lower_err, median+4*upper_err, 0.01)
        fig = plt.figure()
        for i in range(3): plt.axvline([median-lower_err, median, median+upper_err][i], color='k', lw=2)
        plt.plot( x, skewnorm.pdf(x, alpha, loc=sigma, scale=omega), 'r-', lw=2 )
        fit_percentiles = skewnorm.ppf([0.16, 0.5, 0.84], alpha, loc=sigma, scale=omega)
        for i in range(3): plt.axvline(fit_percentiles[i], color='r', ls='--', lw=2)
        plt.hist(samples, density=True, color='red', alpha=0.5)
        return samples, fig



def calculate_skewed_normal_params(median, lower_err, upper_err):
    '''
    Fits a screwed normal distribution via its CDF to the [16,50,84]-percentiles
    
    Inputs:
    -------
    median : float
        the median value that was reported
    lower_err : float
        the lower errorbar that was reported
    upper_err : float
        the upper errorbar that was reported
    size : int
        the number of samples to be drawn
        
    Returns:
    --------
    sigma : float
        the mean of the fitted skewed normal distribution
    omega : float
        the std of the fitted skewed normal distribution
    alpha : float
        the skewness parameter
    '''
    
    lower_err = np.abs(lower_err)
    upper_err = np.abs(upper_err)
    
    def fake_lnlike(p):
        sigma, omega, alpha = p
        eq1 = skewnorm.ppf(0.5, alpha, loc=sigma, scale=omega) - median
        eq2 = skewnorm.ppf(0.16, alpha, loc=sigma, scale=omega) - (median-lower_err)
        eq3 = skewnorm.ppf(0.84, alpha, loc=sigma, scale=omega) - (median+upper_err)
        fake_lnlike = np.log( eq1**2 + eq2**2 + eq3**2 )
#        print fake_lnlike
        return fake_lnlike

    std = np.mean([lower_err, upper_err])
    initial_guess = (median, std, 0) #sigma, omega, alpha 
#    print 'initial_guess:', initial_guess
    sol = minimize(fake_lnlike, initial_guess, bounds=[(None,None), (0,None), (None,None)])
    sigma, omega, alpha = sol.x
#    print 'fake_lnlike:', fake_lnlike((sigma, omega, alpha))
#    print 'Parameters:'
#    print 'sigma', sigma
#    print 'omega', omega
#    print 'alpha', alpha
    
    return sigma, omega, alpha

    
    
if __name__ == '__main__':
    '''
    For testing, simulate a skewed normal distribution with parameters
    alpha_0, loc=sigma_0, scale=omega_0,
    perform the fit and compare the resulting PDF
    '''
    
    ###########################################################################
    # ::: simulate a skewed normal PDF, measure median and errors, and fit
    ###########################################################################
#    sigma_0 = 0.
#    omega_0 = 5.
#    alpha_0 = 2
#    
#    x = np.arange(-20+sigma_0,20+sigma_0,0.01)
#    y = skewnorm.pdf(x, alpha_0, loc=sigma_0, scale=omega_0)
#    plt.figure()
#    plt.plot( x, y, 'b-', lw=4 )
#    
#    median = skewnorm.ppf(0.5, alpha_0, loc=sigma_0, scale=omega_0)
#    lower_err = median - skewnorm.ppf(0.16, alpha_0, loc=sigma_0, scale=omega_0)
#    upper_err = skewnorm.ppf(0.84, alpha_0, loc=sigma_0, scale=omega_0) - median
#    plt.axvline(median)
#    plt.axvline(median-lower_err)
#    plt.axvline(median+upper_err)
#    
#    sigma, omega, alpha = calculate_skewed_normal_params(median, lower_err, upper_err)
#    plt.plot( x, skewnorm.pdf(x, alpha, loc=sigma, scale=omega), 'r--', lw=2 )



    ###########################################################################
    # ::: example inclination
    ###########################################################################
#    (median, lower_err, upper_err) = (84.3, -2.0, 1.3)
#    samples, fig = simulate_posterior_samples(median, lower_err, upper_err, size=1, plot=True)
#    print(np.percentile(samples, [16,50,84]))


#    lower_err = np.abs(lower_err)
#    upper_err = np.abs(upper_err)
#    
#    x = np.arange(median-4*lower_err, median+4*upper_err, 0.01)
#    plt.figure()
#    plt.axvline(median, color='k', lw=2)
#    plt.axvline(median-lower_err, color='k', lw=2)
#    plt.axvline(median+upper_err, color='k', lw=2)
#    
#    sigma, omega, alpha = calculate_skewed_normal_params(median, lower_err, upper_err)
#    plt.plot( x, skewnorm.pdf(x, alpha, loc=sigma, scale=omega), 'r-', lw=2 )
#    
#    fit_percentiles = skewnorm.ppf([0.16, 0.5, 0.84], alpha, loc=sigma, scale=omega)
#    for i in range(3): plt.axvline(fit_percentiles[i], color='r', ls='--', lw=2)
#    
#    rvs = simulate_posterior_samples(median, lower_err, upper_err, size=1000)
#    plt.hist(rvs, density=True, color='lightblue')
